/*
 * dh6.c    	Routines for DHCPv6 data structures
 *
 * Version:     @(#)dh6.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include "dh6.h"

/* Solicit message length excluding extensions */
inline int
dh6_solicit_msg_len(struct dh6_solicit *msg)
{
	return sizeof(struct dh6_solicit) +
		(msg->Abit+msg->Lbit)*IPV6_ADDR_LEN;
}

/* Link-local address if present */
inline struct in6_addr * 
get_solicit_llocal_addr( struct dh6_solicit *msg )
{
	if ( msg->Lbit )
		return( msg->addr );
	else
		return NULL;
}

/* set Link-local address */
inline void
set_solicit_llocal_addr( struct dh6_solicit *msg, struct in6_addr *lladdr )
{
	if ( !msg->Lbit )
		return;
	memcpy(msg->addr, lladdr, IPV6_ADDR_LEN);
}

/* relay agent address if present */
inline struct in6_addr * 
get_solicit_relay_addr( struct dh6_solicit *msg )
{
	if ( msg->Abit )
		return( msg->addr + msg->Lbit );
	else
		return NULL;
}

/* relay agent address if present */
inline void
set_solicit_relay_addr( struct dh6_solicit *msg, struct in6_addr *agent )
{
	if ( !msg->Abit )
		return;
	memcpy(msg->addr+msg->Lbit, agent, IPV6_ADDR_LEN);
}

/* extensions */
inline u_int8_t * 
solicit_ext( struct dh6_solicit *msg )
{
	return( (u_int8_t *) (msg->addr + msg->Lbit + msg->Abit) );
}


/* Advertise message length excluding extensions */
inline int
dh6_advertise_msg_len(struct dh6_advertise *msg)
{
	return sizeof(struct dh6_advertise) +
		(msg->Lbit + msg->SvrCnt)*IPV6_ADDR_LEN;
}

/* Link-local address if present */
inline struct in6_addr * 
get_advertise_llocal_addr( struct dh6_advertise *msg )
{
	if ( msg->Lbit )
		return( msg->addr );
	else
		return NULL;
}

/* set Link-local address */
inline void
set_advertise_llocal_addr( 
	struct dh6_advertise *msg, 
	struct in6_addr *lladdr)
{
	if ( !msg->Lbit )
		return;
	memcpy(msg->addr, lladdr, IPV6_ADDR_LEN);
}

/*  a number of server addresses */
inline struct in6_addr * 
get_advertise_server_addr( struct dh6_advertise *msg )
{
	if (msg->SvrCnt > 0)
		return( msg->addr + msg->Lbit );
	else
		return NULL;
}

/*  a number of server addresses */
inline void
set_advertise_server_addr( 
	struct dh6_advertise *msg,
	struct in6_addr *svr_addr,
	int svr_cnt)
{
	int i;
	if (msg->SvrCnt == 0)
		return;
	for (i=0; i<svr_cnt; i++)
		memcpy(msg->addr+msg->Lbit+i, svr_addr+i, IPV6_ADDR_LEN);
}

/* extensions */
inline u_int8_t * 
advertise_ext( struct dh6_advertise *msg )
{
	return( (u_int8_t *) (msg->addr + msg->Lbit + msg->SvrCnt) );
}


/* Request message length excluding extensions */
inline int
dh6_request_msg_len(struct dh6_request *msg)
{
	return sizeof(struct dh6_request) +
		(msg->Sbit+2)*IPV6_ADDR_LEN;
}

/* server address if present */
inline struct in6_addr * 
get_request_server_addr( struct dh6_request *msg )
{
	if ( msg->Sbit )
		return( msg->addr );
	else
		return NULL;
}

/* server address if present */
inline void
set_request_server_addr( struct dh6_request *msg, struct in6_addr *server )
{
	if ( !msg->Sbit )
		return;
	memcpy(msg->addr, server, IPV6_ADDR_LEN);
}

/* agent address */
inline struct in6_addr * 
get_request_agent_addr( struct dh6_request *msg )
{
	return( msg->addr + msg->Sbit );
}

/* agent address */
inline void
set_request_agent_addr( struct dh6_request *msg, struct in6_addr *agent )
{
	memcpy(msg->addr+msg->Sbit, agent, IPV6_ADDR_LEN);
}

/* Link-local address */
inline struct in6_addr * 
get_request_llocal_addr( struct dh6_request *msg )
{
	return( msg->addr + msg->Sbit + 1 );
}

/* Link-local address */
inline void
set_request_llocal_addr( struct dh6_request *msg, struct in6_addr *lladdr )
{
	memcpy(msg->addr+msg->Sbit+1, lladdr, IPV6_ADDR_LEN);
}

/* extensions */
inline u_int8_t * 
request_ext( struct dh6_request *msg )
{
	return( (u_int8_t *) (msg->addr + msg->Sbit + 2) );
}


/* Reply message length excluding extensions */
inline int
dh6_reply_msg_len(struct dh6_reply *msg)
{
	return sizeof(struct dh6_reply) + msg->Lbit*IPV6_ADDR_LEN;
}

/* Link-local address */
inline struct in6_addr * 
get_reply_llocal_addr( struct dh6_reply *msg )
{
	if ( msg->Lbit )
		return( msg->addr );
	else
		return NULL;
}

/* Link-local address */
inline void
set_reply_llocal_addr( struct dh6_reply *msg, struct in6_addr *lladdr )
{
	if ( !msg->Lbit )
		return;
	memcpy(msg->addr, lladdr, IPV6_ADDR_LEN);
}

/* extensions */
inline u_int8_t * 
reply_ext( struct dh6_reply *msg )
{
	return( (u_int8_t *) (msg->addr + msg->Lbit) );
}

/* Release message length excluding extensions */
inline int
dh6_release_msg_len(struct dh6_release *msg)
{
	return sizeof(struct dh6_release);
}

/* Reconfigure message length excluding extensions */
inline int
dh6_reconfigure_msg_len(struct dh6_reconfigure *msg)
{
	return sizeof(struct dh6_reconfigure);
}

static struct in6_addr 	__dh6sin_agent_group,
			__dh6sin_server_group,
			__dh6sin_relay_group;

static inline void init_dh6_agent_group( void )
{
	inet_pton(AF_INET6, DH6_AGENT_GROUP,  &__dh6sin_agent_group);
}

inline struct in6_addr *
dh6sin_agent_group( void )
{
	static int flag = 0;
	if ( !flag ) {
		init_dh6_agent_group();
		flag = 1;
	}
	return &__dh6sin_agent_group;
}

static inline void init_dh6_server_group( void )
{
	inet_pton(AF_INET6, DH6_SERVER_GROUP, &__dh6sin_server_group);
}

inline struct in6_addr *
dh6sin_server_group( void )
{
	static int flag = 0;
	if ( !flag ) {
		init_dh6_server_group();
		flag = 1;
	}
	return &__dh6sin_server_group;
}

static inline struct in6_addr * 
init_dh6_relay_group( void )
{
	inet_pton(AF_INET6, DH6_RELAY_GROUP,  &__dh6sin_relay_group);
}

inline struct in6_addr *
dh6sin_relay_group( void )
{
	static int flag = 0;
	if ( !flag ) {
		init_dh6_relay_group();
		flag = 1;
	}
	return &__dh6sin_relay_group;
}
