/*
 * dh6.h    	DHCPv6 Data Structures
 *
 * Version:     @(#)dh6.h   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#ifndef __DH6_H
#define __DH6_H

#define DH6_CLIENT_PORT	 546		/* DHCPv6 Client UDP port */
#define DH6_AGENT_PORT	 547		/* DHCPv6 Server/Relay UDP port */

#define DH6_AGENT_GROUP	 "FF02::1:2"	/* DHCPv6 Agent multicast group */
#define DH6_SERVER_GROUP "FF05::1:3"	/* DHCPv6 Server multicast group */
#define DH6_RELAY_GROUP	 "FF05::1:4"	/* DHCPv6 Relay multicast group */

#ifdef DH6_RELEASE

#define DH6_LIFETIME_UNIT 	3600
#define DH6_DEFAULT_PREFER_TIME	24
#define DH6_MAX_PREFER_TIME	240
#define DH6_DEFAULT_VALID_TIME	36
#define DH6_MAX_VALID_TIME	360

#else

#define DH6_LIFETIME_UNIT 	60
#define DH6_DEFAULT_PREFER_TIME	1
#define DH6_MAX_PREFER_TIME	10
#define DH6_DEFAULT_VALID_TIME	2
#define DH6_MAX_VALID_TIME	20

#endif

extern struct in6_addr * dh6sin_agent_group( void );
extern struct in6_addr * dh6sin_server_group( void );
extern struct in6_addr * dh6sin_relay_group( void );

/*
 *	Retransmission and configuration parameters
 */

#define REPLY_MSG_INITIAL_TIMEOUT	2
#define REPLY_MSG_MIN_RETRANS		10
#define REPLY_MSG_RETRANS_INTERVAL	2

#define RECONF_MSG_INITIAL_TIMEOUT	2
#define RECONF_MSG_MIN_RETRANS		10
#define RECONF_MSG_RETRANS_INTERVAL	2
#define RECONF_MSG_MIN_RESP		2
#define RECONF_MSG_MAX_RESP		10

#define RELAY_DISCOVERY_PERIOD		3600

#define MAX_SOLICIT_DELAY		5
#define MAX_ADV_WAIT			5
#define MAX_SOLICIT_RETRANS		10

#define XID_TIMEOUT			10800

#define IPV6_ADDR_LEN			16

/*
 *	DHCP Solicit message
 *	It is safe to allocate a 9x4-byte of buffer for the message
 */

struct dh6_solicit {
#define DH6_SOLICIT	1
	u_int8_t	type;		/* message type */
	u_int8_t	rsvd1:4,	
			Pbit:1,		/* P for Previously cached server */
			Abit:1,		/* A for relay Agent's address */
			Lbit:1,		/* L for Link-local address */
			Cbit:1;		/* C for dellocating resources */
	u_int16_t	rsvd2;
	struct in6_addr	addr[0];	
};

/* Solicit message length excluding extensions */
extern int dh6_solicit_msg_len(struct dh6_solicit *);

/* Link-local address if present */
extern struct in6_addr * get_solicit_llocal_addr(struct dh6_solicit *);
/* set Link-local address */
extern void set_solicit_llocal_addr(struct dh6_solicit *, struct in6_addr *);

/* relay agent address if present */
extern struct in6_addr * get_solicit_relay_addr(struct dh6_solicit *);
/* set relay agent address if present */
extern void set_solicit_relay_addr(struct dh6_solicit *, struct in6_addr *);

/* extensions */
extern u_int8_t * 
solicit_ext( struct dh6_solicit * );

/*
 *	DHCP Advertise message
 */

struct dh6_advertise {
#define DH6_ADVERTISE	2
	u_int8_t	type;		/* message type */
	u_int8_t	rsvd:6,	
			Lbit:1,		/* L for Link-local address */
			Sbit:1;		/* S for Server addresses */
	u_int8_t	SvrCnt;		/* server count */
#define DH6_SVR_TIMEUNIT	4096
#define DH6_NOLIFETIME		0
	u_int8_t	lifetime;	/* lifetime, in units of 4096 seconds */
	struct in6_addr	AgtAddr;	/* agent address */
	struct in6_addr	addr[0];	
};

/* Advertise message length excluding extensions */
extern int dh6_advertise_msg_len(struct dh6_advertise *);

/* Link-local address if present */
extern struct in6_addr * get_advertise_llocal_addr(struct dh6_advertise *);
/* set Link-local address */
extern void set_advertise_llocal_addr(struct dh6_advertise *, struct in6_addr *); 

/*  a number of server addresses */
extern struct in6_addr * get_advertise_server_addr(struct dh6_advertise *);
/*  set a number of server addresses */
extern void set_advertise_server_addr(struct dh6_advertise *, struct in6_addr *,int);

/* extensions */
extern u_int8_t * advertise_ext(struct dh6_advertise *);


/*
 *	DHCP Request message
 */

struct dh6_request {
#define DH6_REQUEST  	3
	u_int8_t	type;		/* message type */
	u_int8_t	rsvd:6,	
			Cbit:1,		/* C for Cleaning up resources */
			Sbit:1;		/* S for Server addresses */
	u_int16_t	id;		/* transaction ID */
	struct in6_addr	addr[0];	
};

/* Request message length excluding extensions */
extern int dh6_request_msg_len(struct dh6_request *);

/* server address if present */
extern struct in6_addr * get_request_server_addr(struct dh6_request *);
/* set server address if present */
extern void set_request_server_addr(struct dh6_request *, struct in6_addr *);

/* agent address */
extern struct in6_addr * get_request_agent_addr(struct dh6_request *);
/* set agent address */
extern void set_request_agent_addr(struct dh6_request *, struct in6_addr *);

/* Link-local address */
extern struct in6_addr * get_request_llocal_addr(struct dh6_request *);
/* set Link-local address */
extern void set_request_llocal_addr(struct dh6_request *, struct in6_addr *);

/* extensions */
extern u_int8_t * request_ext(struct dh6_request *);


/*
 *	DHCP Reply message
 */

struct dh6_reply {
#define DH6_REPLY    	4
	u_int8_t	type;		/* message type */
	u_int8_t	code:7,		/* error code */
			Lbit:1;		/* L for Link-local address */
	u_int16_t	id;		/* transaction ID */
	struct in6_addr	addr[0];	
};

/* Reply message length excluding extensions */
extern int dh6_reply_msg_len(struct dh6_reply *);

/* get Link-local address */
extern struct in6_addr * get_reply_llocal_addr(struct dh6_reply *);
/* set Link-local address */
extern void set_reply_llocal_addr(struct dh6_reply *, struct in6_addr *);

/* extensions */
extern u_int8_t * reply_ext(struct dh6_reply *);


#define DH6_REPLY_SUCCESS        	0
#define DH6_REPLY_FAIL_UNKNOWN   	16
#define DH6_REPLY_AUTH_FAIL      	17
#define DH6_REPLY_POOR_FORM      	18
#define DH6_REPLY_RSRC_UNAVAIL   	19
#define DH6_REPLY_CLI_REC_UNAVAIL	20
#define DH6_REPLY_INVAL_REL_SRC  	21
#define DH6_REPLY_UNABLE_EXT_PARA	22
#define DH6_REPLY_BAD_HAIR_DAY   	24
#define DH6_REPLY_INSUFF_FUND    	32
#define DH6_REPLY_SVR_UNREACH    	64

/*
 *	DHCP Release message
 */

struct dh6_release {
#define DH6_RELEASE  	5
	u_int8_t	type;		/* message type */
	u_int8_t	flags:7,	/* message flags: 0 */
			Dbit:1;		/* D for Directly to the client */
	u_int16_t	id;		/* transaction ID */
	struct in6_addr	AgtAddr;	/* agent address */
	struct in6_addr	lladdr; 	/* Link-local address */
	u_int8_t	ext[0];		/* extensions */
};

/* Release message length excluding extensions */
extern int dh6_release_msg_len(struct dh6_release *);

/*
 *	DHCP Reconfigure message
 */

struct dh6_reconfigure {
#define DH6_RECONFIGURE	6
	u_int8_t	type;		/* message type */
	u_int8_t	flags;		/* message flags: 0 */
	u_int16_t	rsvd;
	u_int8_t	ext[0];		/* extensions */
};

/* Reconfigure message length excluding extensions */
extern int dh6_reconfigure_msg_len(struct dh6_reconfigure *);

#endif /* __DH6_H */
