/*
 * dh6_cli.c    DHCPv6 Client Daemon
 *
 * Version:     @(#)dh6_cli.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include <db.h>
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dispatch.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_if.h"
#include "dh6_cli.h"
#include "dh6_adm.h"
#include "dh6_db.h"

extern void cli_run_unix_server( void * );
extern void cli_adm_msgrcv( void * );
extern void cli_dh6_msg_rcv( void * );
static void cli_dh6_multicast_rcv( void * );
extern void cia_more_reply_timeout( void * );
extern void cia_release_timeout( void * );
extern void cia_prefer_life_timeout( void * );
extern void cia_valid_life_timeout( void * );

extern void cli_solicit_timeout( void * );
extern void cia_request_timeout( void * );
extern int cli_multicast_solicit(char *, ...);
extern int cli_request_send( struct cli_ifaddr *, char *, ...);

extern UNION_LIST 	ifi_table;
UNION_LIST	 server_list,
		 relay_list,
		 client_list;	/* link local address not on this list */

struct u_timer	 solicit_timer;	/* Solicit retransmission timer */
u_int8_t	 solicit_rc;	/* Solicit retransmission counter */
u_int8_t	 send_request_wait = 0;

IF_INFO		*this_ifi;	/* the current interface */

struct fd_list   multi_sock;    /* multicast socket */
struct fd_list	 ll_sock;	/* socket for the link local address */
struct in6_addr	*lladdr;	/* link local address */

extern int	 quit_flag;
extern int	 debug;

int	SolicitPbit = 1;
int	SolicitCbit = 1;
int	RequestLife = DH6_DEFAULT_PREFER_TIME;
int	RequestQbit = 0;

/*
 *	DHCPv6 client main module, called when
 *	1. Router Advertisement comes up with M flag set; OR
 *	2. no router is available after reasonable time; OR
 *	3. in required to adminsitration.
 *
 *	There is one UDP socket for the link-local address or each 
 *	DHCPv6 address. The link-local socket should be always opened.
 *	After the client acquires an intended DHCPv6 address, 
 *	the socket should be opened accordingly.
 */

int dh6_client_main( int ifindex )
{
	struct fd_list   ufd;
	struct poll_list upl;
	struct cli_ifaddr *cia;
        int     optval;
     
        u_free_list_init();
	reset_timer_table();
	reset_fd_list();
	reset_poll_list();

	begin_timing();
	init_ifi_table();
	get_ifi_table();

	dh6_client_init();

	if ( cli_check_ifi_valid(ifindex) < 0 ) {
		log_warn(1101, "interface invalid");
		exit(1);
	}

	/*
	 *	build up interface extension;
	 *	build up Link-local address extension;
	 * 	and etc;
	 */

	/*
	 *	reboot from previous cache;
	 *	restore all the lists;
	 * 	and etc.
	 */
	read_dh6_client_conf();	/* read from "dh6_client.conf" */
	read_dh6_client_boot();	/* read from "dh6_client.boot" */

        /* open unix socket; listen for incoming administration commands. */
        if ( (ufd.fd = open_unix_sock(
	      dh6cli_adm_server_name(ifi_ifname(this_ifi)))) < 0 )
                log_fatal(1102, "open unix socket fail");

	init_poll(&upl);
	upl.data = &ufd;
	upl.function = cli_run_unix_server;
	add_poll(&upl);

        if ( (multi_sock.fd=cli_udp6_socket(NULL)) < 0)
                log_fatal(1103, "open UDP6 socket fail");

#ifdef IPV6_TXINFO_WORK
        optval = 1;
        if (setsockopt(multi_sock.fd, IPPROTO_IPV6, IPV6_TXINFO,
                        (char *)&optval, sizeof(optval)) < 0)
                log_fatal(1104, "IPV6_TXINFO fail");
#endif  
                
        optval = 1;
        if (setsockopt(multi_sock.fd, IPPROTO_IPV6, IPV6_RXINFO,
                        (char *)&optval, sizeof(optval)) < 0)
                log_fatal(1105, "IPV6_RXINFO fail");
        
        init_fdp(&multi_sock);
        multi_sock.data = &multi_sock;
        multi_sock.function = cli_dh6_multicast_rcv;
        add_fdp(&multi_sock);

#ifdef DH6_DEBUG
	if ( look_xmit_more() )
		log_info("bind to link local address %s", 
			inet_ntop(AF_INET6, lladdr, 16, NULL));
#endif
	/* open the link-local socket */
	if ( (ll_sock.fd=cli_udp6_socket(lladdr)) < 0 ) 
                log_fatal(1103, "open UDP6 socket fail");

        if (setsockopt(ll_sock.fd, IPPROTO_IPV6, IP_MULTICAST_IF,
			(char *)lladdr, IPV6_ADDR_LEN) < 0)
                log_fatal(1104, "IP_MULTICAST_IF fail");

	/*
	 * 	TODO: we should do something here to set the
	 *	      hop limit to 1.
	 */

        init_fdp(&ll_sock);
        ll_sock.data = NULL;
        ll_sock.function = cli_dh6_msg_rcv;
        add_fdp(&ll_sock);

	/* open all the sockets for DHCPv6 address */
	for ( cia = client_list.cia_head; cia != NULL; cia = cia->next ) {
		if ( cia_addr_verified(cia) ) {
	 		if ((cia->cia_sk.fd=cli_udp6_socket(&cia->cia_addr)) 
					< 0)
                		log_fatal(1105, "open UDP6 socket fail");

		        if (setsockopt(cia->cia_sk.fd, IPPROTO_IPV6,
				IP_MULTICAST_IF, (char *)&cia->cia_addr, 
				IPV6_ADDR_LEN) < 0)
		                log_fatal(1106, "IP_MULTICAST_IF fail");
		
        		init_fdp(&cia->cia_sk);
        		cia->cia_sk.data = cia;
        		cia->cia_sk.function = cli_dh6_msg_rcv;
        		add_fdp(&cia->cia_sk);
		}
	}

	init_timer(&solicit_timer);
	solicit_timer.data = (void *)NULL;
	solicit_timer.function = cli_solicit_timeout;

	/*
	 *	if there is no server on list, start relay/server discovery 
	 *	procedure using the link-local socket. Otherwise, 
	 *	initiate Request/Reply transaction.
	 */
	if ( server_list.cs_head == NULL ) {
        	cli_multicast_solicit("PC", SolicitPbit, SolicitCbit);
        	/* 
		 * TODO: a random delay between 0 and MAX_SOLICIT_DELAY 
		 *	 see more in section 5.1, draft 08
		 */
        	start_timer(&solicit_timer, MAX_ADV_WAIT*SECONDS);
        	solicit_rc = 1;
		send_request_wait = 1;
	}
	else {
		cia = client_list.cia_head;
		if ( cia == NULL ) 	
        		cli_start_request_proc(NULL, NULL, NULL, 0, 0, NULL);
		else {
			for ( ; cia != NULL; cia = cia->next ) 
        			cli_start_request_proc(cia, NULL, NULL, 1,
					0, NULL);
		}
		send_request_wait = 0;
	}

        quit_flag = 0;
        for (;;) {
                dispatch();
                if ( quit_flag )
                        break;
        }

	end_timing();
	del_poll(&upl);
	for ( cia = client_list.cia_head; 
		  cia != NULL; cia = cia->next ) {
		if ( cia->cia_sk.fd >= 0 ) {
			del_fdp( &cia->cia_sk );
			close( cia->cia_sk.fd );
		}
	}
	del_fdp( &ll_sock ); close( ll_sock.fd );
        del_fdp(&multi_sock); close(multi_sock.fd);
	close_unix_sock( ufd.fd, dh6cli_adm_server_name(ifi_ifname(this_ifi)) );
	closelog();
	exit(0);
}

inline struct cli_relay * cr_create( void )
{
	struct cli_relay *cr;
    	if ( (cr=u_alloc_node()) == NULL )
                return NULL;
	cr->cr_issvr = NULL;
        return cr;
}

inline void cr_destroy( struct cli_relay *cr )
{
	if ( cr->cr_issvr ) {
		cr->cr_issvr->cs_rly = NULL;
		/* the server entry should be released separately */
	}
        u_free_node(cr);
}

inline struct cli_relay * cr_get( struct in6_addr *addr )
{
	struct cli_relay *cr;
	for (cr = relay_list.cr_head; cr != NULL;
		cr = cr->next) {
		if ( !memcmp(&cr->cr_addr, addr, IPV6_ADDR_LEN) )
			return cr;
	}
	return NULL;
}

inline struct cli_server * cs_create( void )
{
	struct cli_server *cs;
    	if ( (cs=u_alloc_node()) == NULL )
                return NULL;
	cs->cs_rly = NULL;
        return cs;
}

inline void cs_destroy( struct cli_server *cs )
{
        u_free_node(cs);
}

inline struct cli_server * cs_get( struct in6_addr *addr )
{
	struct cli_server *cs;
	for (cs = server_list.cs_head; cs != NULL;
		cs = cs->next) {
		if ( !memcmp(&cs->cs_addr, addr, IPV6_ADDR_LEN) )
			return cs;
	}
	return NULL;
}

struct cli_ifaddr * cia_create( void )
{
	struct cli_ifaddr *cia;
    	if ( (cia=u_alloc_node()) == NULL )
                return NULL;

	cia->cia_ifi = NULL;
	cia->cia_svr = NULL;
	cia->cia_rly = NULL;

	init_timer(&cia->cia_xmit_timer);
	cia->cia_xmit_timer.data = (void *)cia;
        cia->cia_xmit_timer.function =
                        (void (*)(void *))cia_request_timeout;
	cia->cia_xmit_rc = 0;

	init_timer(&cia->cia_life_timer);
	cia->cia_life_timer.data = (void *)cia;
	cia->cia_life_timer.function = u_ignore;

	cia->cia_state = CIA_MSG_NONE | CIA_ADDR_NONE;
	cia->cia_sk.fd = 0;
	cia->cia_extlen = 0;
	cia->cia_ext = NULL;

        return cia;
}

inline void cia_destroy( struct cli_ifaddr *cia )
{
	close( cia->cia_sk.fd );
	stop_timer(&cia->cia_xmit_timer);
	stop_timer(&cia->cia_life_timer);
	if ( cia->cia_ext )
		free(cia->cia_ext);
        u_free_node(cia);
}

inline struct cli_ifaddr * cia_get( struct in6_addr *addr )
{
	struct cli_ifaddr *cia;
	for (cia = client_list.cia_head; cia != NULL;
		cia = cia->next) {
		if ( !memcmp(&cia->cia_addr, addr, IPV6_ADDR_LEN) )
			return cia;
	}
	return NULL;
}

inline struct cli_ifaddr * 
cia_get_by_server_id( struct in6_addr *server, int id )
{
	struct cli_ifaddr *cia;
	for (cia = client_list.cia_head; cia != NULL;
		cia = cia->next) {
		if ( !memcmp(&cia->cia_svr->cs_addr, server, IPV6_ADDR_LEN)
		     && cia->cia_id == id )
			return cia;
	}
	return NULL;
}

inline struct cli_ifaddr * 
cia_get_by_relay_id( struct in6_addr *relay, int id )
{
	struct cli_ifaddr *cia;
	for (cia = client_list.cia_head; cia != NULL;
		cia = cia->next) {
		if ( !memcmp(&cia->cia_rly->cr_addr, relay, IPV6_ADDR_LEN)
		     && cia->cia_id == id )
			return cia;
	}
	return NULL;
}

inline struct cli_ifaddr * 
cia_get_by_server_relay( struct in6_addr *server, struct in6_addr *relay )
{
	struct cli_ifaddr *cia;
	for (cia = client_list.cia_head; cia != NULL;
		cia = cia->next) {
		if ( !memcmp(&cia->cia_svr->cs_addr, server, IPV6_ADDR_LEN)
		  && !memcmp(&cia->cia_rly->cr_addr, relay,  IPV6_ADDR_LEN) )
			return cia;
	}
	return NULL;
}

static int cia_next_trans_id = 0;

static inline int cia_get_next_trans_id( void )
{
	return cia_next_trans_id++;
}

int dh6_client_init( void )
{
	server_list.chain.head = NULL;
	server_list.chain.tail = NULL;
	relay_list.chain.head  = NULL;
	relay_list.chain.tail  = NULL;
	client_list.chain.head = NULL;
	client_list.chain.tail = NULL;
}

/*
 *	Validation of interface and DHCPv6 address
 *	To be completed ...
 */

int cli_check_ifi_valid( int ifindex )
{
	this_ifi = seek_ifi_byindex(ifindex);
	if (this_ifi == NULL)
		return -1;
	lladdr = ifi_lladdr(this_ifi);

	return 1;
}

/*
 *  to receive DHCPv6 multicast messages
 */
void cli_dh6_multicast_rcv( void * data )
{
	struct fd_list *sock = (struct fd_list *)data;
    	u_int8_t buf[MAX_MSG_LEN];
	u_int8_t ctrldata[64];

    	for (;;) {
		struct iovec aiov[1];
		struct msghdr msg;
		struct cmsghdr *cmsg = (struct cmsghdr *)ctrldata;
		DH6_MESSAGE    *dh6msg = (DH6_MESSAGE *)buf;
		struct sockaddr_in6 sin6;
    		int  len;

		msg.msg_name = (caddr_t)&sin6;
		msg.msg_namelen = sizeof(struct sockaddr_in6) ;
		msg.msg_iov = aiov;
		msg.msg_iovlen = 1;
		aiov[0].iov_base = buf;
		aiov[0].iov_len = MAX_MSG_LEN;
		msg.msg_control = ctrldata;
		msg.msg_controllen = sizeof(ctrldata);
		msg.msg_flags = 0;
		len = recvmsg(sock->fd, &msg, 0);

        	/*
         	 *  It is very possible for signal SIGCHLD to interrupt this
         	 *  slow system call recvfrom
         	 */
__ugly0:
	        if (len == 0)
	            	return;
	        else if (len < 0) {
	            	if (errno == EINTR)
	                	goto __ugly0;
                        if (errno == EWOULDBLOCK) 
                                return;  /* no more messages */
	            	log_sys(1114, "recvfrom");
	        }

	        switch ( dh6msg->msg.type ) {
	            	case DH6_RECONFIGURE: 
				cli_reconfigure_rcv(NULL, &msg, len);
				break;
	            	default:
	                	log_warn(3015,"unknown message %u", 
					dh6msg->msg.type);
	        }
	}
	return;
}

/*
 *  to receive DHCPv6 messages
 */
void cli_dh6_msg_rcv( void * data )
{
	struct cli_ifaddr *cia = (struct cli_ifaddr *)data;
	int fd = (cia) ? cia->cia_sk.fd : ll_sock.fd;
        u_int8_t buf[MAX_MSG_LEN];
        u_int8_t ctrldata[64];

      	for (;;) {
                struct iovec aiov[1];
                struct msghdr msg;
                struct cmsghdr *cmsg = (struct cmsghdr *)ctrldata;
                DH6_MESSAGE    *dh6msg = (DH6_MESSAGE *)buf;
                struct sockaddr_in6 sin6;
                int  len;
        
                msg.msg_name = (caddr_t)&sin6;
                msg.msg_namelen = sizeof(struct sockaddr_in6) ;
                msg.msg_iov = aiov;
                msg.msg_iovlen = 1;
                aiov[0].iov_base = buf;
                aiov[0].iov_len = MAX_MSG_LEN;
                msg.msg_control = ctrldata;
                msg.msg_controllen = sizeof(ctrldata);
                msg.msg_flags = 0;
                len = recvmsg(fd, &msg, 0);

                /*
                 *  It is very possible for signal SIGCHLD to interrupt this
                 *  slow system call recvfrom
                 */
__ugly:
                if (len == 0)
                        return;
                else if (len < 0) {
                        if (errno == EINTR)
                                goto __ugly;
                        if (errno == EWOULDBLOCK)
                                return;  /* no more messages */
                        log_sys(1107, "recvmsg");
                }

	        switch ( dh6msg->msg.type ) {
	            	case DH6_ADVERTISE:	  
				cli_advertise_rcv(cia, &msg, len);
				break;
	            	case DH6_REPLY: 	  
				cli_reply_rcv(cia, &msg, len);
				break;
	            	case DH6_RECONFIGURE: 
				cli_reconfigure_rcv(cia, &msg, len);
				break;
	            	default:
	                	log_warn(1108,"unknown message %u", 
					dh6msg->msg.type);
	        }
	}
	return;
}

/*
 *	On receiving DHC Advertise message
 */
int cli_advertise_rcv(struct cli_ifaddr *cia, struct msghdr *msg, int len)
{
        struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
	struct cli_relay	*cr;
	struct cli_server	*cs;
	struct in6_addr		*addr;
	int i;
        
#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("receiving Advertise message from %s",
                        inet_ntop(AF_INET6, &sin6->sin6_addr, 16, NULL));
#endif          

	if (dh6msg->adv.SvrCnt == 0 && dh6msg->adv.Sbit == 0) {
		log_warn(1109, "SvrCnt==0 but Sbit==0");
		return -1;
	}

	if (dh6msg->adv.Sbit && IN6_IS_MULTICAST(&dh6msg->adv.AgtAddr) ) {
		log_warn(1110, "Found multicast server address");
		return -1;
	}
	addr = get_advertise_server_addr(&dh6msg->adv);
	for (i=0; i<dh6msg->adv.SvrCnt; i++) {
		if ( IN6_IS_MULTICAST(addr) ) {
			log_warn(1111, "Found multicast server address");
			return -1;
		}
		addr++;
	}

	if ( !(cr=cr_get(&dh6msg->adv.AgtAddr)) ) {
		if ( (cr=cr_create()) == NULL )
			return -1;
		memcpy(&cr->cr_addr, &dh6msg->adv.AgtAddr, IPV6_ADDR_LEN);
		u_queue_head(&relay_list, cr);

		if ( dh6msg->adv.Sbit ) {	/* also server address */
			if ( (cs=cs_create()) == NULL )
				return -1;
			memcpy(&cs->cs_addr, &dh6msg->adv.AgtAddr,
					IPV6_ADDR_LEN);
			cs->cs_rly = cr;
			u_queue_head(&server_list, cs);
			cr->cr_issvr = cs;
		}
	}

	addr = get_advertise_server_addr(&dh6msg->adv);
	for (i=0; i<dh6msg->adv.SvrCnt; i++) {
		if ( !(cs=cs_get(addr)) ) {
			if ( (cs=cs_create()) == NULL )
				return -1;
			memcpy(&cs->cs_addr, addr, IPV6_ADDR_LEN);
			u_queue_head(&server_list, cs);
		}
		cs->cs_rly = cr;
		addr++;
	}
	stop_timer(&solicit_timer);

	if ( send_request_wait ) {
		struct cli_ifaddr *cia2;
		if ( client_list.cia_head == NULL ) 	
        		cli_start_request_proc(NULL, NULL, NULL, 0, 0, NULL);
		else {
			for (cia2 = client_list.cia_head; cia2 != NULL; 
					cia2 = cia2->next ) 
        			cli_start_request_proc(cia2, NULL, NULL, 1, 
					0, NULL);
		}
		send_request_wait = 0;
	}

	/*
	 *	TODO: Am I interested in extensions if any?
	 */
	return 0;
}

/*
 *	On receiving DHC Reply message
 */
int cli_reply_rcv(struct cli_ifaddr *cia, struct msghdr *msg, int len)
{
	struct cli_ifaddr *cia2 = NULL;
        struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
	struct ext6_addr *addrext;
	struct ext6_multicast *mcastext;
	DB *db;
                        
#ifdef DH6_DEBUG          
        if ( look_xmit() )
                log_info("receiving Reply message from %s",
                        inet_ntop(AF_INET6, &sin6->sin6_addr, 16, NULL));
#endif  

	if (dh6msg->rpl.Lbit &&
	    (cia2=cia_get_by_relay_id(&sin6->sin6_addr,ntohs(dh6msg->rpl.id))) 
			== NULL ) {
		log_warn(1112, "Reply not forwarded by relay");
		return -1;
	}
	if (!dh6msg->rpl.Lbit &&
	    (cia2=cia_get_by_server_id(&sin6->sin6_addr,ntohs(dh6msg->rpl.id))) 
			== NULL ) {
		log_warn(1113, "Reply not sent by server directly");
		return -1;
	}

	addrext = (struct ext6_addr *)ext6_get(EXT6_ADDR, 
			reply_ext(&dh6msg->rpl), 
			len - dh6_reply_msg_len(&dh6msg->rpl));
	mcastext = (struct ext6_multicast *)ext6_get(EXT6_MULTICAST, 
			reply_ext(&dh6msg->rpl), 
			len - dh6_reply_msg_len(&dh6msg->rpl));
	if (addrext == NULL) {
		log_warn(1115, "No address extension in Reply message");
		return -1;
	}
	if ( !addrext->Cbit ) {
		log_warn(1116, "No client address found in Reply extension");
		return -1;
	}

#ifdef TODO
	/* to check if it provides the correct authetication info. */
	if ( fail_authen_info(&dh6msg->rpl) ) {
    		cli_request_send(cia, "");
       		start_timer(&cia->cia_xmit_timer, REPLY_MSG_RETRANS_INTERVAL);
       		cia->cia_xmit_rc++;
		log_warn("fail in providing the correct authetication info.");
		return -1;
	}
#endif

	/* in respond to Release message */
	if ( cia2->cia_state & CIA_REL_WAIT ) {
#ifdef TODO
                void *ext2 = NULL;
                int   len2 = 0; 
                for (;;) {      
                        ext2 = server_not_released(&dh6msg->rpl,&len2);
                        if (ext2 == NULL || len2 == 0)
                                break;
                        log_info("server not release error %d", 
						dh6msg->rpl.code);
                }
                for (;;) {
                        ext2 = server_released(&dh6msg->rpl,&len2);
                        if (ext2 == NULL || len2 == 0)
                                break;
                        release_particular_parameters(cia,msg);
                        cli_purge_boot(cia, msg);
                }
		if ( some_extensions_pending(cia,msg) )
			return 0; 	/* to wait for more */
#endif
		stop_timer(&cia2->cia_xmit_timer);
		cia2->cia_state |= CIA_REL_CLRD;

#ifdef TODO
		if ( all_parameters_released(cia,msg) ) {
#endif
			dh6_remove_addr6(ifi_ifname(this_ifi), 
			 	strlen(ifi_ifname(this_ifi)), 
			 	&cia2->cia_addr);

		        if ( (db=dh6db_open(DH6CLI_BOOT)) < 0)
                		log_sys(1117, "dbopen");
        		if (dh6db_locate(db, cia2) == 0) {
        			dh6db_del(db, cia2, R_CURSOR);
				dh6db_sync(db);
			}
        		dh6db_close(db);

			u_unlink(&client_list, cia2);
			cia_destroy(cia2);
#ifdef TODO
		}
#endif
		return 0;
	}

	/* in respond to Request message */
	if ( cia2->cia_xmit_rc == 0 &&
		/* received the replies for all the retransmissions */
	     !(cia2->cia_state & CIA_REQ_WAIT ||
	       cia2->cia_state & CIA_RPL_MORE) ) {
		log_warn(1118,"Reply received in an incorrect state");
		return -1;
	}

	if (dh6msg->rpl.code != DH6_REPLY_SUCCESS) {
		stop_timer(&cia2->cia_xmit_timer);
		log_info("Request rejected with code %d", dh6msg->rpl.code);
#ifdef TODO
		if ( some_extension_missing( reply_ext(dh6msg->rpl) ) ) {
   			cia2->cia_xmit_timer.function = 
				(void (*)(void *))cia_more_reply_timeout;
			start_timer(&cia2->cia_xmit_timer, 
				    REPLY_MSG_INITIAL_TIMEOUT);
			/* to receive more Reply */
			return 0;
		}
#endif
		if ( !(cia2->cia_state & CIA_ADDR_PRFD ||
		       cia2->cia_state & CIA_ADDR_VALID) ) {
			u_unlink(&client_list, cia2);
			cia_destroy(cia2);
		}
		else
			cia2->cia_state |= CIA_REQ_CLRD;
		return 0;
	}

	cia2->cia_prfsize = addrext->PrfSize;
	memcpy( &cia2->cia_addr, 
	        get_ext6_addr_client_addr(addrext), 
		IPV6_ADDR_LEN );

	cia2->cia_prefer_life = get_ext6_addr_prefer_lifetime(addrext);
	cia2->cia_valid_life  = get_ext6_addr_valid_lifetime(addrext);
	stop_timer(&cia2->cia_xmit_timer);
	if (cia2->cia_xmit_rc > 0)
		cia2->cia_xmit_rc -= 1;
	cia2->cia_req_life -= get_ext6_addr_prefer_lifetime(
				(struct ext6_addr *)reply_ext(&dh6msg->rpl));

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("Got a DHCPv6 address %s!!!",
			inet_ntop(AF_INET6, &cia2->cia_addr, 16, NULL));
#endif
#ifdef TODO
	if ( the_first_time_assigned(&cia2->cia_addr) )
		dup_addr_detect(&cia2->cia_addr);
#endif
	dh6_assign_addr6(ifi_ifname(this_ifi), 
			 strlen(ifi_ifname(this_ifi)), 
			 &cia2->cia_addr, 
			 cia2->cia_prfsize, 0);

        if ( (db=dh6db_open(DH6CLI_BOOT)) < 0)
                log_sys(1119, "dbopen");
        dh6db_put(db, cia2, R_NOOVERWRITE);
	dh6db_sync(db);
        dh6db_close(db);

        cia2->cia_life_timer.function =
                        (void (*)(void *))cia_prefer_life_timeout;
        start_timer(&cia2->cia_life_timer, 
		cia2->cia_prefer_life*DH6_LIFETIME_UNIT);
        cia2->cia_state &= CIA_ADDR_PRFD | 0x0f;

	if ((cia2->cia_sk.fd=cli_udp6_socket(&cia2->cia_addr)) 
			< 0)
               	log_fatal(1120, "open UDP6 socket fail");

        if (setsockopt(cia2->cia_sk.fd, IPPROTO_IPV6,
		IP_MULTICAST_IF, (char *)&cia2->cia_addr, 
		IPV6_ADDR_LEN) < 0)
                log_fatal(1121, "IP_MULTICAST_IF fail");
		
    	init_fdp(&cia2->cia_sk);
      	cia2->cia_sk.data = cia2;
      	cia2->cia_sk.function = cli_dh6_msg_rcv;
       	add_fdp(&cia2->cia_sk);

        /* join server multicast group */
	if (mcastext != NULL) {
	        struct ipv6_mreq mr6;
        	int     optval;

                memcpy(&mr6.ipv6mr_interface, &cia2->cia_addr, IPV6_ADDR_LEN);
                memcpy(&mr6.ipv6mr_multiaddr, &mcastext->addr, IPV6_ADDR_LEN);
                if (setsockopt(cia2->cia_sk.fd, IPPROTO_IPV6, IP_ADD_MEMBERSHIP,
                                                (caddr_t)&mr6, sizeof(mr6)) < 0)
                        log_fatal(3007, "server IP_ADD_MEMBERSHIP fail");
	}

#ifdef TODO
	/*
	 * 1. to handle other extension;
	 * 2. to check if any missing extensions;
	 */
	cli_process_other_extensions( reply_ext(dh6msg->rpl) );
	if ( some_extension_missing( reply_ext(dh6msg->rpl) ) ) {
   		cia2->cia_xmit_timer.function = 
			(void (*)(void *))cia_more_reply_timeout;
		start_timer(&cia2->cia_xmit_timer, REPLY_MSG_INITIAL_TIMEOUT);
		/* to receive more Reply */
		return 0;
	}
#endif
	cia2->cia_state |= CIA_REQ_CLRD;

	return 0;
}

/*
 *	On receiving DHC Reconfigure message
 */
int cli_reconfigure_rcv(struct cli_ifaddr *cia, struct msghdr *msg, int len)
{
        struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
	int	 extlen = 0;
	void 	*ext = NULL;
	struct cli_ifaddr *cia2 = cia;
                        
#ifdef DH6_DEBUG          
        if ( look_xmit() )
                log_info("receiving Reconfigure message from %s",
                        inet_ntop(AF_INET6, &sin6->sin6_addr, 16, NULL));
#endif  

	if ( cia2 == NULL ) {
		if ((cia2=client_list.cia_head) == NULL) {
			log_warn(1116, "no client entry");
			return -1;
		}
	}

	if (memcmp(&sin6->sin6_addr, &cia2->cia_svr->cs_addr, IPV6_ADDR_LEN)) {
        	cli_start_request_proc(NULL, NULL, NULL, 0, 0, NULL);
		return 0;
	}

#ifdef TODO
	while ( extract_an_extension(&dh6msg->conf) ) {
		append_matching_extension(&ext, &extlen);
	}
#endif
    	cia2->cia_xmit_rc = 0;
	cia2->cia_id = cia_get_next_trans_id();
	/*
	 * 	should delay a random amount of time between
	 *	RECONF_MSG_MIN_RESP and RECONF_MSG_MAX_RESP
	 */
    	cli_start_request_proc(cia2, &cia2->cia_svr->cs_addr, 
		&cia2->cia_rly->cr_addr, 1, extlen, ext);

	return 0;
}


/*
 *	Start to send DHCP Request message
 */
int cli_start_request_proc( 
	struct cli_ifaddr *cia,		/* client entry */ 
	struct cli_server *cs, 		/* server entry */
	struct cli_relay  *cr,		/* relay entry */
	int                Cbit,	/* Clean bit */
	int                extlen,	/* length of extension */
	void		  *ext)
{
	struct cli_ifaddr *cia2 = cia;
	int fd;

#ifdef DH6_DEBUG          
        if ( look_xmit_more() )
                log_info("cli_start_request_proc(cia=%p,cs=%p,cr=%p,Cbit=%d)", 
			cia, cs, cr, Cbit);
#endif  

	if ( cia2 == NULL ) {
		if ( server_list.cs_head == NULL ) {
			/* server not available */
			log_warn(1123, "server not available");
			return -1;
		}

		if ( (cia2=cia_create()) == NULL )
			return -1;
		cia2->cia_ifi = this_ifi;
		u_queue_tail(&client_list, cia2);
	}
	cia2->cia_svr = (cs) ? cs : server_list.cs_head;
	cia2->cia_rly = (cr) ? cr : cia2->cia_svr->cs_rly;
   	cia2->cia_xmit_rc = 0;
	cia2->cia_id = cia_get_next_trans_id();
	fd = ( cia_addr_verified(cia2) ) ? cia2->cia_sk.fd : ll_sock.fd;

	if (extlen == 0 || ext == NULL) {
		cia2->cia_extlen = 0;
		cia2->cia_ext = CLI_APPEND_ADDR_EXT6(cia2->cia_ext,
					&cia2->cia_extlen,
					RequestLife,
					DH6_DEFAULT_VALID_TIME, 
					RequestQbit);
	} else {
		cia2->cia_extlen = extlen;
		cia2->cia_ext = ext;
	}

	if ( cia2->cia_rly->cr_issvr ) {
		cli_request_send(cia2, "SCdf", 
			0, Cbit, &cia2->cia_svr->cs_addr, fd);
	} else {
		cli_request_send(cia2, "SCdf", 
			1, Cbit, &cia2->cia_rly->cr_addr, fd);
	}
    	start_timer(&cia2->cia_xmit_timer, REPLY_MSG_RETRANS_INTERVAL);
    	cia2->cia_xmit_rc++;
    	cia2->cia_state |= CIA_REQ_WAIT;
	return 0;
}


void cli_run_unix_server( void * data )
{
        struct fd_list          *p = (struct fd_list *)data, *fdp;
        int                     fd;
        struct sockaddr_un      unix_addr;
        int                     unix_len = sizeof(unix_addr);

        for (;;) {
                fd = accept(p->fd, (struct sockaddr *)&unix_addr, &unix_len);
                if (fd < 0) {
                        if ( errno!=EWOULDBLOCK )
                                log_sys(1124, "accept");
                        else
                                return;  /* no more pending connections */
                } else {
                        if ( (fdp=malloc(sizeof(struct fd_list))) == NULL )
                                log_sys(1125, "malloc");
                        init_fdp(fdp);
                        fdp->fd   = fd;
                        fdp->data = fdp;
                        fdp->function = cli_adm_msgrcv;
                        add_fdp(fdp);
                }
        }
}

/*
 *	to receive administration commands
 */
void cli_adm_msgrcv( void * data )
{
        struct fd_list *p = (struct fd_list *)data;
        ADM_MESSAGE  msg;
        DH6_MESSAGE *dh6=(DH6_MESSAGE *)(&msg.un.dh6.msg);
        int len;

        for (;;) {
                len = recv(p->fd, (char*)&msg, MAX_MSG_LEN, 0);
                if ( len==0 )
                        return;
                else if (len < 0) {
                        if (errno == EINTR)
                                continue;
                        else
                                log_sys(1126, "recv");
                }

#ifdef DH6_DEBUG
        	if ( look_xmit() )
                        log_info("receiving a message from dh6cc");
#endif
                if (msg.cmd == ADM_CMD_STOP ) {
                        quit_flag = 1;
                        break;
                }

                if (msg.cmd == ADM_SHOW_DB ) {
			cli_dbshow_adm(msg, len);
			continue;
		}
                if (msg.cmd == ADM_DH6_MSG ) {
                	switch ( dh6->msg.type ) {
	            	case DH6_REQUEST:     
				cli_request_adm(&msg, len);       
				break;
	            	case DH6_RELEASE:     
				cli_release_adm(&msg, len);       
				break;
                    	default:
                        	log_warn(1128,"unknown message %u", 
					dh6->msg.type);
			}
                }

        }
        close(p->fd); p->fd=-1;
        del_fdp(p);
        free(p);
}

/*
 *	On receiving a Request message from another process
 */
int cli_request_adm( ADM_MESSAGE *msg, int len )
{
        DH6_MESSAGE *dh6=(DH6_MESSAGE *)(&msg->un.dh6.msg);
	struct cli_ifaddr *cia;
	struct cli_server *cs = NULL;
	struct cli_relay  *cr = NULL;
	struct sockaddr_in6  to6;
        int fd;

	if ( dh6->req.Sbit ) 
		cs = cs_get(get_request_server_addr(&dh6->req));
	else
		cs = cs_get(&msg->un.dh6.faddr);
	if (cs == NULL) {
		log_warn(1129, "server address not on list");
		return -1;
	}
	cr = cr_get(get_request_agent_addr(&dh6->req));
	if (cr == NULL) {
		log_warn(1129, "relay address not on list");
		return -1;
	}

	cia = cia_get_by_server_relay(&cs->cs_addr, &cr->cr_addr);
        if (cia == NULL) {   
                if ( (cia=cia_create()) == NULL )
                        return -1;
                cia->cia_ifi = this_ifi;
                u_queue_tail(&client_list, cia);
        }                       
        cia->cia_svr = cs;
        cia->cia_rly = cr;
        cia->cia_id = cia_get_next_trans_id();
 
        cia->cia_extlen = msg->len - dh6_request_msg_len(&dh6->req) 
				   - ADM_DH6MSG_HDRLEN;
        if ( cia->cia_ext )
                cia->cia_ext = realloc(cia->cia_ext, cia->cia_extlen);
        else
                cia->cia_ext = malloc(cia->cia_extlen);
	memcpy(cia->cia_ext, request_ext(&dh6->req), cia->cia_extlen);
 
    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, &msg->un.dh6.faddr, sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_AGENT_PORT);
	dh6->req.id = htons(cia->cia_id);

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Request len:%d, to: %s, port:%d", 
			msg->len-ADM_DH6MSG_HDRLEN,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

        fd = ( cia_addr_verified(cia) ) ? cia->cia_sk.fd : ll_sock.fd;
_resend_request2:
	if (sendto(fd, dh6, msg->len-ADM_DH6MSG_HDRLEN, 0, 
		(struct sockaddr *)&to6, sizeof(struct sockaddr_in6)) 
			!= msg->len-ADM_DH6MSG_HDRLEN) {
	        if ( errno==EINTR )
	                goto _resend_request2;
	        if ( errno==EINVAL ) 
	                goto _resend_request2;
		log_sys_warn(1136, "sendto");
	}

        cia->cia_xmit_timer.function =
                        (void (*)(void *))cia_request_timeout;
        start_timer(&cia->cia_xmit_timer, REPLY_MSG_RETRANS_INTERVAL);
        cia->cia_xmit_rc = 1;
        cia->cia_state |= CIA_REQ_WAIT;
	return 0;
}

/*
 *	On receiving a Release message from another process
 */
int cli_release_adm( ADM_MESSAGE *msg, int len )
{
        DH6_MESSAGE *dh6=(DH6_MESSAGE *)(&msg->un.dh6.msg);
        struct cli_ifaddr *cia;
        struct cli_server *cs = NULL;
        struct cli_relay  *cr = NULL;
        struct sockaddr_in6  to6;  
        int fd;

	cia = cia_get_by_server_relay(&msg->un.dh6.faddr, &dh6->rel.AgtAddr);
	if ( cia == NULL ) {
		log_warn(1129, "The client entry to be released not found");
		return -1;
	}
	if (msg->len-ADM_DH6MSG_HDRLEN <= dh6_release_msg_len(&dh6->rel)) { 
			/* no extension */
		cli_start_request_proc(cia, NULL, NULL, 1, 0, NULL);
		return 0;
	}

	cia->cia_id = cia_get_next_trans_id();
        cia->cia_extlen = msg->len - dh6_release_msg_len(&dh6->rel)
				   - ADM_DH6MSG_HDRLEN;
        if ( cia->cia_ext )
                cia->cia_ext = realloc(cia->cia_ext, cia->cia_extlen);
        else
                cia->cia_ext = malloc(cia->cia_extlen);
	memcpy(cia->cia_ext, dh6->rel.ext, cia->cia_extlen);
 
    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, &msg->un.dh6.faddr, sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_AGENT_PORT);
	dh6->req.id = htons(cia->cia_id);

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Release len:%d, to: %s, port:%d", 
			msg->len-ADM_DH6MSG_HDRLEN,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

        fd = ( cia_addr_verified(cia) ) ? cia->cia_sk.fd : ll_sock.fd;
_resend_release2:
	if (sendto(fd, dh6, msg->len-ADM_DH6MSG_HDRLEN, 0, 
		(struct sockaddr *)&to6, sizeof(struct sockaddr_in6)) 
			!= msg->len-ADM_DH6MSG_HDRLEN) {
	        if ( errno==EINTR )
	                goto _resend_release2;
	        if ( errno==EINVAL )
	                goto _resend_release2;
		log_sys_warn(1137, "sendto");
	}

        cia->cia_xmit_timer.function =
                        (void (*)(void *))cia_release_timeout;
    	start_timer(&cia->cia_xmit_timer, REPLY_MSG_RETRANS_INTERVAL);
        cia->cia_xmit_rc = 1;
	cia->cia_state |= CIA_REL_WAIT;
	return 0;
}

/*
 *	Solicit transmission timeout
 */
void cli_solicit_timeout( void * data )
{
	if ( solicit_rc <= MAX_SOLICIT_RETRANS ) {
    		cli_multicast_solicit("");
        	start_timer(&solicit_timer, MAX_ADV_WAIT*SECONDS);
        	solicit_rc++;
	}
	else
		log_warn(1129,"agent not available");
}

/*
 *	Request transmission timeout
 */
void cia_request_timeout( void * data )
{
	struct cli_ifaddr	*cia = (struct cli_ifaddr *)data;

	if ( cia->cia_xmit_rc <= REPLY_MSG_MIN_RETRANS ) {
    		cli_request_send(cia, "");
        	start_timer(&cia->cia_xmit_timer, REPLY_MSG_RETRANS_INTERVAL);
        	cia->cia_xmit_rc++;
	}
	else {
		struct cli_server *cs;
		for (cs = server_list.cs_head; cs != NULL;
				cs = cs->next) {
			if (cs != cia->cia_svr) {
				cia->cia_svr = cs;
				cia->cia_rly = cs->cs_rly;
				break;
			}
		}
		if (cs != NULL) {
			if ( !cs->cs_rly->cr_issvr ) 
    				cli_request_send(cia, 
					"Sd", 1, &cia->cia_rly->cr_addr);
			else
    				cli_request_send(cia, 
					"Sd", 0, &cs->cs_addr);
        		start_timer(&cia->cia_xmit_timer, 
				REPLY_MSG_RETRANS_INTERVAL);
        		cia->cia_xmit_rc = 1;
		} else
			log_warn(1130,"server not available");
	}
}

/*
 *	Timeout for receiving more reply 
 */
void cia_more_reply_timeout( void * data )
{
	struct cli_ifaddr	*cia = (struct cli_ifaddr *)data;

       	cia->cia_xmit_timer.function =
                        (void (*)(void *))cia_request_timeout;
    	start_timer(&cia->cia_xmit_timer, REPLY_MSG_RETRANS_INTERVAL);
	cli_request_send(cia, "");
    	cia->cia_xmit_rc++;
}

/*
 *	Timeout for transmitting a Release message
 */
void cia_release_timeout( void * data )
{
	struct cli_ifaddr	*cia = (struct cli_ifaddr *)data;
	DB *db;

	log_warn(1131, "Release not reponded by server");
	cia->cia_state |= CIA_REL_CLRD;

#ifdef TODO
	release_desired_parameters(cia);
	cli_purge_boot(cia);
	if ( all_parameters_tobe_released(cia) ) {
#endif
		dh6_remove_addr6(ifi_ifname(this_ifi), 
		 	strlen(ifi_ifname(this_ifi)), 
		 	&cia->cia_addr);
	        if ( (db=dh6db_open(DH6CLI_BOOT)) < 0)
	              	log_sys(1132, "dbopen");
	       	if (dh6db_locate(db, cia) == 0) {
	       		dh6db_del(db, cia, R_CURSOR);
			dh6db_sync(db);
		}
	     	dh6db_close(db);
	        u_unlink(&client_list, cia);
	        cia_destroy(cia);
#ifdef TODO
	}
#endif
}

/*
 *	When the preferred lifetime expires
 */
void cia_prefer_life_timeout( void * data )
{
	struct cli_ifaddr	*cia = (struct cli_ifaddr *)data;
	DB *db;

	if ( cia->cia_req_life > 0 ) {
		/* to continue to request the same address */
		cia->cia_id = cia_get_next_trans_id();
		cli_request_send(cia, "SCdf", 
			0, 0, &cia->cia_svr->cs_addr, cia->cia_sk.fd);
        	cia->cia_xmit_timer.function =
                        (void (*)(void *))cia_request_timeout;
    		start_timer(&cia->cia_xmit_timer, REPLY_MSG_RETRANS_INTERVAL);
    		cia->cia_xmit_rc = 1;
    		cia->cia_state |= CIA_REQ_WAIT;
	}
        cia->cia_life_timer.function =
                        (void (*)(void *))cia_valid_life_timeout;
        start_timer(&cia->cia_life_timer, 
		(cia->cia_valid_life - cia->cia_prefer_life)*DH6_LIFETIME_UNIT);
	cia->cia_state &= CIA_ADDR_VALID | 0x0f;
        if ( (db=dh6db_open(DH6CLI_BOOT)) < 0)
              	log_sys(1133, "dbopen");
        if (dh6db_locate(db, cia) == 0)
        	dh6db_put(db, cia, R_CURSOR);
	else
        	dh6db_put(db, cia, R_NOOVERWRITE);
	dh6db_sync(db);
     	dh6db_close(db);
}

/*
 *	When the valid lifetime expires
 */
void cia_valid_life_timeout( void * data )
{
	struct cli_ifaddr	*cia = (struct cli_ifaddr *)data;
	DB *db;

	log_info("DHCP address %s becomes invalid", 
		inet_ntop(AF_INET6, &(cia->cia_addr), 16, NULL));
	dh6_remove_addr6(ifi_ifname(this_ifi), 
	 	strlen(ifi_ifname(this_ifi)), 
	 	&cia->cia_addr);
        if ( (db=dh6db_open(DH6CLI_BOOT)) < 0)
              	log_sys(1134, "dbopen");
       	if (dh6db_locate(db, cia) == 0) {
       		dh6db_del(db, cia, R_CURSOR);
		dh6db_sync(db);
	}
     	dh6db_close(db);
        u_unlink(&client_list, cia);
        cia_destroy(cia);
}

int cli_multicast_solicit(char *which, ...)
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;
        va_list 	ap;
	static u_int8_t	Pbit = 1, Cbit = 1;

        va_start(ap, which);
	while ( *which ) {
		switch (*which++) {
		case 'P':
			Pbit = va_arg(ap, u_int8_t); 
			break;
		case 'C':
			Cbit = va_arg(ap, u_int8_t); 
			break;
		}
	}
        va_end(ap);

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, dh6sin_agent_group(), sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_AGENT_PORT);

    	bzero((char *)msg, MAX_MSG_LEN);
	msg->sol.type = DH6_SOLICIT;
	msg->sol.Pbit = Pbit;
	msg->sol.Cbit = Cbit;

	len = sizeof(struct dh6_solicit);

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Solicit len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                        ntohs(to6.sin6_port));
#endif 

_resend_solicit:
	if (sendto(ll_sock.fd, msg, len, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_solicit;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Solicit");
	                goto _resend_solicit;
		}
		log_sys(1135, "sendto");
	}

	/*
	 *	TODO: Should I send any extensions?
	 */
	return 0;
}

int cli_request_send(
	struct cli_ifaddr *cia,		/* client entry */
	char		  *which, 	/* which parameters */
	... )
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;
        va_list 	ap;
	static u_int8_t	Cbit = 1, Sbit = 1;
	static struct in6_addr dest;
	static int	fd = -1; /* you must set "fd" at the first time */

        va_start(ap, which);
	while ( *which ) {
		switch (*which++) {
		case 'S':
			Sbit = va_arg(ap, u_int8_t); 
			break;
		case 'C':
			Cbit = va_arg(ap, u_int8_t); 
			break;
		case 'd':	/* destination address */
			memcpy(&dest, va_arg(ap, struct in6_addr *), 
					IPV6_ADDR_LEN);
			break;
		case 'f':	/* socket */
			fd = va_arg(ap, int); 
			break;
		}
	}
        va_end(ap);

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, &dest, sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_AGENT_PORT);

    	bzero((char *)msg, sizeof(DH6_MESSAGE));
	msg->req.type = DH6_REQUEST;
	msg->req.Sbit = Sbit;
	msg->req.Cbit = Cbit;
	msg->req.id   = htons(cia->cia_id);

	if ( Sbit ) {
		set_request_server_addr(&msg->req, &cia->cia_svr->cs_addr);
		set_request_agent_addr(&msg->req, &cia->cia_rly->cr_addr);
	} else 
		set_request_agent_addr(&msg->req, &cia->cia_svr->cs_addr);
	set_request_llocal_addr(&msg->req, ifi_lladdr(cia->cia_ifi));

	if (cia->cia_extlen > 0)
		memcpy(request_ext(&msg->req), cia->cia_ext, cia->cia_extlen);

	len = sizeof(struct dh6_request) + 
	      IPV6_ADDR_LEN*(msg->req.Sbit+2) +
	      cia->cia_extlen;

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Request len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_request:
	if (sendto(fd, msg, len, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_request;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Request");
	                goto _resend_request;
		}
		log_sys_warn(1136, "sendto");
	}

	return 0;
}

int cli_release_send(
	struct cli_ifaddr *cia,		/* client entry */
	struct in6_addr   *dest,	/* destination address */
	int		   Dbit)	/* Direct or not */
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;
	int		fd = (cia->cia_sk.fd>0) ? cia->cia_sk.fd : ll_sock.fd;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, dest, sizeof(struct in6_addr));
    	to6.sin6_port = htons(DH6_AGENT_PORT);

    	bzero((char *)msg, sizeof(DH6_MESSAGE));
	msg->rel.type = DH6_RELEASE;
	msg->rel.Dbit = Dbit;
	msg->rel.id   = htons(cia->cia_id);
	memcpy(&msg->rel.AgtAddr, &cia->cia_rly->cr_addr, IPV6_ADDR_LEN);
	memcpy(&msg->rel.lladdr, ifi_lladdr(cia->cia_ifi), IPV6_ADDR_LEN);

	if (cia->cia_extlen > 0)
		memcpy(msg->rel.ext, cia->cia_ext, cia->cia_extlen);

	len = dh6_release_msg_len(&msg->rel) + cia->cia_extlen;

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Release len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_release:
	if (sendto(fd, msg, len, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_release;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Release");
	                goto _resend_release;
		}
		log_sys_warn(1137, "sendto");
	}

	return 0;
}
