/*
 * dh6_db.c    	Routines for DHCPv6 Database
 *
 * Version:     @(#)dh6_db.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include <db.h>
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dispatch.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_adm.h"
#include "dh6_db.h"
#include "parse.h"

u_int16_t key_prefix = 0x80FE; /* ntohs(0xFE80) */

inline int dh6db_copy(void *dst, void *src, int len)
{
	u_int8_t  *dst2 = dst, *src2 = src;
	int  i;
	for (i = 0; i < len; i++)
		dst2[i] = src2[i];
}

inline DB * dh6db_open(const char *fname)
{
        char name[100];
	DB   *db;
        sprintf(name, "%s%s", DH6_SERV_PATH, fname);
	db = dbopen(name, O_CREAT | O_RDWR | O_NONBLOCK,
			S_IRUSR | S_IWUSR, DB_RECNO, NULL);
			/* ??? */
#ifdef DH6_DEBUG
        if ( look_db_more() )
		log_info("open database %s fd:%d", name, db->fd(db));
#endif
	return db;
}

inline int dh6db_close(DB *db)
{
#ifdef DH6_DEBUG
        if ( look_db_more() )
		log_info("close database fd:%d", db->fd(db));
#endif
	return db->close(db);
}

inline int dh6db_fd(const DB *db)
{
	return db->fd(db);
}

inline int dh6db_sync(const DB *db)
{
#ifdef DH6_DEBUG
        if ( look_db_more() )
		log_info("sync database fd:%d", db->fd(db));
#endif
	return db->sync(db, 0);
}

int dh6db_del(DB *db, void *entry, u_int flags)
{
	int ret = 0;
	DBT key;
#ifdef DH6_DEBUG
        if ( look_db_more() )
		log_info("delete a record in database fd:%d", db->fd(db));
#endif
	dh6db_alloc_key(&key, entry);
	ret = db->del(db, &key, flags);
	free(key.data);
	return ret;
}

int dh6db_get(DB *db, void *entry, u_int flags)
{
	int ret = 0;
	DBT rec, key;
	dh6db_alloc_key(&key, entry);
	ret = db->get(db, &key, &rec, flags);
	free(key.data);
	if (ret == 0 && rec.size > 0) {
#ifdef DH6_DEBUG
	        if ( look_db_more() )
			log_info("get a record (size:%d) in database fd:%d", 
				rec.size, db->fd(db));
#endif
		dh6db_convert_rec(&rec, entry);
		return 0;
	} else
		return -1;
}

int dh6db_put(DB *db, void *entry, u_int flags)
{
	int ret;
	DBT rec, key;
	dh6db_alloc_rec(&rec, entry);
	dh6db_alloc_key(&key, entry);
#ifdef DH6_DEBUG
        if ( look_db_more() ) 
		log_info("put a record (size:%d) in database fd:%d", 
			rec.size, db->fd(db));
#endif
	ret = db->put(db, &key, &rec, flags);
	free(rec.data);
	free(key.data);
	return ret;
}

int dh6db_locate(DB *db, void *entry)
{
	int ret;
	DBT rec, key;
	dh6db_alloc_key(&key, entry);
	ret = db->seq(db, &key, &rec, R_CURSOR);
	free(key.data);
	if (ret == 0 && rec.size > 0) {
#ifdef DH6_DEBUG
	        if ( look_db_more() )
			log_info("locate a record (size:%d) in database fd:%d", 
				rec.size, db->fd(db));
#endif
		dh6db_convert_rec(&rec, entry);
		return 0;
	} else
		return -1;
}

int dh6db_skip_next(DB *db, void *entry)
{
	int ret;
	DBT rec, key;
	memset(&rec, 0, sizeof(rec));
	memset(&key, 0, sizeof(key));
	ret = db->seq(db, &key, &rec, R_NEXT);
	if (ret == 0 && rec.size > 0) {
#ifdef DH6_DEBUG
        	if ( look_db_more() )
			log_info("skip to next record (size:%d) in db fd:%d", 
				rec.size, db->fd(db));
#endif
		dh6db_convert_rec(&rec, entry);
		return 0;
	}
	else
		return -1;
}

int dh6db_skip_prev(DB *db, void *entry)
{
	int ret;
	DBT rec, key;
	memset(&rec, 0, sizeof(rec));
	memset(&key, 0, sizeof(key));
	ret = db->seq(db, &key, &rec, R_PREV);
	if (ret == 0 && rec.size > 0) {
#ifdef DH6_DEBUG
        	if ( look_db_more() )
			log_info("skip to prev record (size:%d) in db fd:%d", 
				rec.size, db->fd(db));
#endif
		dh6db_convert_rec(&rec, entry);
		return 0;
	} else
		return -1;
}
