/*
 * dh6_svr.c    DHCPv6 Server Daemon
 *
 * Version:     @(#)dh6_svr.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include <db.h>
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dispatch.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_if.h"
#include "dh6_svr.h"
#include "dh6_adm.h"
#include "sock.h"
#include "dh6_db.h"

static void svr_run_unix_server( void * );
static void svr_adm_msgrcv( void * );
static void svr_dh6_msg_rcv( void * );
static void svr_dh6_multicast_rcv( void * );
extern void sc_xid_timeout( void * );
extern void db_check_timeout( void * );

struct fd_list   multi_sock;	/* multicast socket */
struct fd_list   unix_sock;	/* unix socket */
struct poll_list poll_unix;

extern UNION_LIST 	ifi_table;
UNION_LIST	 server_list;	/* link local address not on this list */
struct u_timer	 db_check_timer = /* timer for checking/purging db */
		  	{ NULL, NULL, NULL, 
			  (void (*)(void *))db_check_timeout };

int 	debug = 0;  
int  	quit_flag = 0;
int	MaxPreferLife = DH6_MAX_PREFER_TIME;
int	MaxValidLife  = DH6_MAX_VALID_TIME;
struct in6_addr MulticastAddr;

/*
 *	There is a UDP socket for each server address;
 *	There is only one multicast socket;
 */

void main( int argc, char *argv[] )
{
	struct ipv6_mreq mr6;
        int     optval;
        char    op;
	struct svr_ifaddr *sia;

        opterr = 0;
        while ((op = getopt(argc, argv, "d:D:")) != EOF) {
                switch (op) {
#ifdef DH6_DEBUG
                        case 'd':   /* debug */
                        case 'D':   /* debug */
                                debug = atoi(optarg);
                                break;
#endif
                        default:
                }
        }

        if ( daemon_init() < 0 )
                log_fatal(3001, "daemon init fail");

        if (debug > 0)
                openlog ("", LOG_CONS, LOG_DAEMON);
        else    
                openlog (argv[0], 0, LOG_DAEMON);

        u_free_list_init();
        begin_timing();

        init_ifi_table();
        get_ifi_table();
#ifdef DH6_DEBUG
        if ( look_cache() ) {
                log_info("debug = %d", debug);
                print_ifi_table();
        }
#endif  

	dh6_server_init();
	svr_get_addr_table();
#ifdef DH6_DEBUG
        if ( look_cache() ) 
		svr_print_addr_table();
#endif

        if ( (unix_sock.fd=open_unix_sock(dh6svr_adm_server_name())) < 0 )
                log_fatal(3002, "open unix socket fail");

        init_poll(&poll_unix);
        poll_unix.data = &unix_sock;
        poll_unix.function = svr_run_unix_server;
        add_poll(&poll_unix);

	if ( (multi_sock.fd=svr_udp6_socket(NULL)) < 0)
               	log_fatal(3003, "open UDP6 socket fail");

#ifdef IPV6_TXINFO_WORK
        optval = 1;
        if (setsockopt(multi_sock.fd, IPPROTO_IPV6, IPV6_TXINFO,
			(char *)&optval, sizeof(optval)) < 0)
                log_fatal(3004, "IPV6_TXINFO fail");
#endif

        optval = 1;
        if (setsockopt(multi_sock.fd, IPPROTO_IPV6, IPV6_RXINFO,
			(char *)&optval, sizeof(optval)) < 0)
                log_fatal(3005, "IPV6_RXINFO fail");

	init_fdp(&multi_sock);
	multi_sock.data = &multi_sock;
	multi_sock.function = svr_dh6_multicast_rcv;
	add_fdp(&multi_sock);

	/* open a socket for each server address */
	for (sia = server_list.sia_head;
		sia != NULL; sia = sia->next) {

		if ( (sia->sia_sk.fd=svr_udp6_socket(&sia->sia_addr)) < 0)
                	log_fatal(3006, "open UDP6 socket fail");

		memcpy(&mr6.ipv6mr_interface, &sia->sia_addr, 
			sizeof(struct in6_addr)); 
		
		/* join server multicast group */
		memcpy(&mr6.ipv6mr_multiaddr, dh6sin_server_group(),
					sizeof(struct in6_addr));
		if (setsockopt(sia->sia_sk.fd, IPPROTO_IPV6, IP_ADD_MEMBERSHIP,
						(caddr_t)&mr6, sizeof(mr6)) < 0)
			log_fatal(3007, "server IP_ADD_MEMBERSHIP fail");
		
		/* join agent multicast group */
		memcpy(&mr6.ipv6mr_multiaddr, dh6sin_agent_group(),
						sizeof(struct  in6_addr));
		if (setsockopt(sia->sia_sk.fd, IPPROTO_IPV6, IP_ADD_MEMBERSHIP,
						(caddr_t)&mr6, sizeof(mr6)) < 0)
			log_fatal(3008, "agent IP_ADD_MEMBERSHIP fail");

	        if (setsockopt(sia->sia_sk.fd, IPPROTO_IPV6, IP_MULTICAST_IF,
	       	                (char *)&sia->sia_addr, IPV6_ADDR_LEN) < 0)
	       	        log_fatal(3009, "IP_MULTICAST_IF fail");

		
	        init_fdp(&sia->sia_sk);
	        sia->sia_sk.data = sia;
	        sia->sia_sk.function = svr_dh6_msg_rcv;
	        add_fdp(&sia->sia_sk);
	}

	read_dh6_server_conf();	/* read from "dh6_server.conf" */

	/* 
	 *	start timer to check db once per SVR_DB_CHECK_INTERVAL
	 */
        init_timer(&db_check_timer);
	db_check_timer.data = NULL;
	db_check_timer.function = (void (*)(void *))db_check_timeout;
	start_timer(&db_check_timer, SVR_DB_CHECK_INTERVAL);

        for (;;) {
                dispatch();
                if ( quit_flag )
                        break;
        }

        end_timing();

        del_poll(&poll_unix);
        close_unix_sock(unix_sock.fd, dh6svr_adm_server_name());
	del_fdp(&multi_sock); close(multi_sock.fd);
	for (sia = server_list.sia_head;
		sia != NULL; sia = sia->next) {
		if (sia->sia_sk.fd > 0) {
			del_fdp(&sia->sia_sk);
			close(sia->sia_sk.fd);
		}
	}

	stop_timer(&db_check_timer);
#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("exit from %s", argv[0]);
#endif
        closelog();
	exit(0);
}

struct svr_client * sc_create( void )
{
	struct svr_client *sc;
    	if ( (sc=u_alloc_node()) == NULL )
                return NULL;
	sc->sc_sia = NULL;
	sc->sc_state = SC_STATE_NONE;

        init_timer(&sc->sc_xmit_timer);
        sc->sc_xmit_timer.data = (void *)sc;
        sc->sc_xmit_timer.function = u_ignore;
        sc->sc_xmit_timeout = 0;

        init_timer(&sc->sc_dns_timer);
        sc->sc_dns_timer.data = (void *)sc;
        sc->sc_dns_timer.function = u_ignore;

        return sc;
}

inline void sc_destroy( struct svr_client *sc )
{
	stop_timer(&sc->sc_xmit_timer);
	stop_timer(&sc->sc_dns_timer);
        sc->sc_xmit_timeout = 0;
        u_free_node(sc);
}

struct svr_client * sc_get(
	struct svr_ifaddr	*sia,
	struct in6_addr		*ll_addr,
	struct in6_addr		*rly_addr)
{
	struct svr_client	*sc;
	if (sia == NULL)
		return NULL;
	for (sc = sia->sia_clilist.sc_head;
		sc != NULL; sc = sc->next) {
		if ( !memcmp(&sc->sc_lladdr, ll_addr, IPV6_ADDR_LEN) &&
		     !memcmp(&sc->sc_rlyaddr,rly_addr,IPV6_ADDR_LEN) )
			return sc;
	}
	return NULL;
}

struct svr_ifaddr * sia_create( void )
{
	struct svr_ifaddr *sia;
    	if ( (sia=u_alloc_node()) == NULL )
                return NULL;

	sia->sia_ifi = NULL;

	sia->sia_clilist.sc.head = NULL;
	sia->sia_clilist.sc.tail = NULL;

	sia->sia_sk.fd = -1;
	sia->sia_lifetime = SERVER_DEFAULT_LIFETIME;
	/*
	 *	Server should periodically multicast Advertise messages
	 */
        return sia;
}

inline void sia_destroy( struct svr_ifaddr *sia )
{
	if (sia->sia_sk.fd > 0) {
		del_fdp(&sia->sia_sk);
		close(sia->sia_sk.fd);
	}
        u_free_node(sia);
}

struct svr_ifaddr * sia_get(struct in6_addr *addr)
{
	struct svr_ifaddr *sia;
	for (sia = server_list.sia_head;
		sia != NULL; sia = sia->next) {
		if ( !memcmp(&sia->sia_addr, addr, IPV6_ADDR_LEN) )
			return sia;
	}
	return NULL;
}

int dh6_server_init( void )
{
        server_list.chain.head = NULL;
        server_list.chain.tail = NULL;
	memset(&MulticastAddr, 0, IPV6_ADDR_LEN);
}

int svr_get_addr_table( void )
{
        u_char buf[8192]; 
        u_char *buf_end;
        struct ifconf ifc;
        struct ifreq *ifr;
        struct if_info *ifi;
        struct svr_ifaddr *sia;
        struct sockaddr_in6 *sin6;
	int s;
 
	if ((s=socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
		log_sys(3010, "socket");
        ifc.ifc_len = sizeof(buf);
        ifc.ifc_buf = (char *)buf; 
        if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0)
                log_fatal(3011,"SIOCGIFCONF failed");
        buf_end = buf + ifc.ifc_len;
        
        ifr = ifc.ifc_req;
        while ((u_char *)ifr < buf_end) {
                sin6 = (struct sockaddr_in6 *)&ifr->ifr_addr;
                if ( ifr->ifr_addr.sa_family == AF_INET6 &&
                     !IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr) ) {
                        if ( (sia = u_alloc_node()) == NULL )
                                return -1;
                        memset(sia, 0, sizeof(struct svr_ifaddr));
                        sia->sia_ifi = seek_ifi_byname(ifr->ifr_name,
                                                strlen(ifr->ifr_name));
                        memcpy(&sia->sia_addr, &sin6->sin6_addr,
                                sizeof(struct in6_addr));
                        sia->sia_lifetime = SERVER_DEFAULT_LIFETIME;
                        sia->sia_clilist.sc.head = NULL;
                        sia->sia_clilist.sc.tail = NULL;
                        u_queue_tail(&server_list, sia);
                }
                ifr = (struct ifreq *)((char *)ifr + sizeof(ifr->ifr_name) +
                        (sizeof(ifr->ifr_addr) > ifr->ifr_addr.sa_len ?
                         sizeof(ifr->ifr_addr) : ifr->ifr_addr.sa_len ));
        }
	close(s);
        return 0;
}

#ifdef DH6_DEBUG
int svr_print_addr_table( void )
{
	struct svr_ifaddr *sia;
	int n = 1;

	log_info("Server table:");
        for (sia = server_list.sia_head;
                sia != NULL; sia = sia->next) {
		log_info("    addr: %s, if: %s, lifetime: %d", 
				inet_ntop(AF_INET6,&sia->sia_addr,16,NULL),
				ifi_ifname(sia->sia_ifi), sia->sia_lifetime);
	}
	return 0;
}
#endif

struct svr_ifaddr * svr_get_server_byifname(char *name, int len)
{
	struct svr_ifaddr 	*sia;
	for (sia = server_list.sia_head;
		sia != NULL; sia = sia->next) {
		if ( !memcmp(ifi_ifname(sia->sia_ifi), name, len) )
			return sia;
	}
	return NULL;
}

struct svr_ifaddr * svr_get_server_byifindex(int ifindex)
{
	struct svr_ifaddr 	*sia;
	for (sia = server_list.sia_head;
		sia != NULL; sia = sia->next) {
		if (ifi_ifindex(sia->sia_ifi) == ifindex)
			return sia;
	}
	return NULL;
}

/*
 *	Check the server database and purge record if necessarily
 */
static int dh6db_check_purge( void )
{
	DB  *db, *db2;
	DBT key, rec;
	struct svr_record *p;
	u_int32_t prefer, valid;

        if ( (db=dh6db_open(DH6_SVR_INUSE)) == NULL)
                log_sys(3012, "dbopen error");
       	if ( (db2=dh6db_open(DH6_SVR_INVAL)) == NULL)
              	log_sys(3013, "dbopen error");

        memset(&key, 0, sizeof(key));
	key.size = KPREFIX_LEN;
	key.data = &key_prefix;
	db->seq(db, &key, &rec, R_CURSOR);

	do {
		if (rec.size < sizeof(struct svr_record)) 
			break;
#ifdef DH6_DEBUG
		if ( look_db_more() )
			log_info("Got a next record (size:%d)in purging",
					rec.size);
#endif
		/* remove from DH6_SVR_INUSE and save to DH6_SVR_INVAL */
		p = (struct svr_record *)rec.data;
		dh6db_copy(&prefer, &p->prefer_life, 4);
		dh6db_copy(&valid, &p->valid_life, 4);
		if (valid == 0) {
#ifdef DH6_DEBUG
			if ( look_db_more() )
				log_info("deleting a record");
#endif
			db->del(db, &key, R_CURSOR);
			db->sync(db, R_RECNOSYNC);
			p->state = SC_STATE_DELETE;
			db2->put(db2, &key, &rec, R_NOOVERWRITE);
			db2->sync(db2, R_RECNOSYNC);
		}

		/* change the state */
		else if (prefer == 0) {
#ifdef DH6_DEBUG
			if ( look_db_more() )
				log_info("changing the state of a record");
#endif
			p->state = SC_STATE_VALID;
			db->put(db, &key, &rec, R_NOOVERWRITE);
			db->sync(db, R_RECNOSYNC);
		}
		/*
		 * TODO: Never delete or change the state of 
	 	 *       permanent client records.
		 */
	} while (db->seq(db, &key, &rec, R_NEXT) == 0);

        dh6db_close(db2);
        dh6db_close(db);

	return 0;
}

/*
 *  to receive DHCPv6 messages
 */
static void svr_dh6_multicast_rcv( void * data )
{
	struct fd_list *sock = (struct fd_list *)data;
    	u_int8_t buf[MAX_MSG_LEN];
	u_int8_t ctrldata[64];

    	for (;;) {
		struct iovec aiov[1];
		struct msghdr msg;
		struct cmsghdr *cmsg = (struct cmsghdr *)ctrldata;
		DH6_MESSAGE    *dh6msg = (DH6_MESSAGE *)buf;
		struct sockaddr_in6 sin6;
    		int  len;

		msg.msg_name = (caddr_t)&sin6;
		msg.msg_namelen = sizeof(struct sockaddr_in6) ;
		msg.msg_iov = aiov;
		msg.msg_iovlen = 1;
		aiov[0].iov_base = buf;
		aiov[0].iov_len = MAX_MSG_LEN;
		msg.msg_control = ctrldata;
		msg.msg_controllen = sizeof(ctrldata);
		msg.msg_flags = 0;
		len = recvmsg(sock->fd, &msg, 0);

        	/*
         	 *  It is very possible for signal SIGCHLD to interrupt this
         	 *  slow system call recvfrom
         	 */
__ugly:
	        if (len == 0)
	            	return;
	        else if (len < 0) {
	            	if (errno == EINTR)
	                	goto __ugly;
                        if (errno == EWOULDBLOCK) 
                                return;  /* no more messages */
	            	log_sys(3014, "recvfrom");
	        }

	        switch ( dh6msg->msg.type ) {
	            	case DH6_SOLICIT:	  
				svr_solicit_rcv(&msg, len);
				break;
	            	default:
	                	log_warn(3015,"unknown message %u", 
					dh6msg->msg.type);
	        }
	}
	return;
}


/*
 *	On receiving DHC Solicit message
 */
int svr_solicit_rcv(struct msghdr *msg, int len)
{
	struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
	DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
	struct in6_pktinfo *pktinfo;
	struct svr_ifaddr  *sia;
	int	 extlen = 0;
	void 	*ext = NULL;

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("receiving Solicit message from %s",
			inet_ntop(AF_INET6, &src_addr->sin6_addr, 16, NULL));
#endif
#ifdef TODO
	if ( dest_is_server_group(&dh6msg->sol) &&
	     IN6_IS_ADDR_LINKLOCAL(&src_addr->sin6_addr) ) {
		log_warn(3016, "src is link-local but dst is server multicast");
		return -1;
	}
#endif
	if ( !memcmp(&src_addr->sin6_addr, dh6sin_server_group(), 
		sizeof(struct in6_addr)) &&
	     IN6_IS_ADDR_LINKLOCAL(&src_addr->sin6_addr) ) {
		log_warn(3017, "src addr is link-local but from a relay");
		return -1;
	}
	if ( dh6msg->sol.Lbit && !dh6msg->sol.Abit ){
		log_warn(3018, "malfunctionaing relay agent");
		return -1;
	}
	if (len != sizeof(struct dh6_solicit) 
			+ dh6msg->sol.Lbit * IPV6_ADDR_LEN
			+ dh6msg->sol.Abit * IPV6_ADDR_LEN ) {
		log_warn(3019, "Solicit message length incorrect");
		return -1;
	}

	if ((pktinfo=cmsg_get(msg, IPPROTO_IPV6, IPV6_RXINFO)) == NULL) {
		log_warn(3020, "IPV6_RXINFO not found");
		return -1;
	}

	if ((sia=svr_get_server_byifindex(pktinfo->ipi6_ifindex)) == NULL) {
		log_warn(3021, "server address entry cannot be located");
		return -1;
	}

#ifdef TODO
	svr_alloc_advertise_ext(&ext, &extlen);
#endif
	if ( dh6msg->sol.Lbit )
		svr_advertise_send(sia, 
			get_solicit_llocal_addr(&dh6msg->sol),
			get_solicit_relay_addr(&dh6msg->sol),
			extlen, ext);
	else
		svr_advertise_send(sia, &src_addr->sin6_addr, NULL,extlen,ext);
#ifdef TODO
	free(ext);
#endif

	/* 
	 *	Cleaning up the database for the specific client.
	 */
	if ( dh6msg->sol.Cbit ) {
	        DB  *db, *db2;
		DBT key, rec;
		struct svr_record *p;

        	if ( (db=dh6db_open(DH6_SVR_INUSE)) == NULL)
                	log_sys(3022, "dbopen error");
        	if ( (db2=dh6db_open(DH6_SVR_INVAL)) == NULL)
        	        log_sys(3023, "dbopen error");

        	memset(&key, 0, sizeof(key));
		key.size = KPREFIX_LEN;
		key.data = &key_prefix;
		db->seq(db, &key, &rec, R_CURSOR);

		do {
			if (rec.size < sizeof(struct svr_record)) 
				break;
			p = (struct svr_record *)rec.data;
			if ( !memcmp(get_solicit_llocal_addr(&dh6msg->sol),
				    &p->lladdr, IPV6_ADDR_LEN) ) {
                        	db->del(db, &key, R_CURSOR);
                        	db->sync(db, R_RECNOSYNC);
                        	db2->put(db2, &key, &rec, R_NOOVERWRITE);
                        	db2->sync(db2, R_RECNOSYNC);
			}
                } while (db->seq(db, &key, &rec, R_NEXT) == 0);

        	dh6db_close(db2);
        	dh6db_close(db);
	}

#ifdef TODO
	look_into_other_extensions(&dh6msg->sol);
#endif
	return 0;
}

/*
 *  to receive DHCPv6 messages
 */
static void svr_dh6_msg_rcv( void * data )
{
	struct svr_ifaddr *sia = (struct svr_ifaddr *)data;
    	u_int8_t buf[MAX_MSG_LEN];
	u_int8_t ctrldata[64];

    	for (;;) {
		struct iovec aiov[1];
		struct msghdr msg;
		struct cmsghdr *cmsg = (struct cmsghdr *)ctrldata;
		DH6_MESSAGE    *dh6msg = (DH6_MESSAGE *)buf;
		struct sockaddr_in6 sin6;
    		int  len;

		msg.msg_name = (caddr_t)&sin6;
		msg.msg_namelen = sizeof(struct sockaddr_in6) ;
		msg.msg_iov = aiov;
		msg.msg_iovlen = 1;
		aiov[0].iov_base = buf;
		aiov[0].iov_len = MAX_MSG_LEN;
		msg.msg_control = ctrldata;
		msg.msg_controllen = sizeof(ctrldata);
		msg.msg_flags = 0;
		len = recvmsg(sia->sia_sk.fd, &msg, 0);

        	/*
         	 *  It is very possible for signal SIGCHLD to interrupt this
         	 *  slow system call recvfrom
         	 */
__ugly2:
	        if (len == 0)
	            	return;
	        else if (len < 0) {
	            	if (errno == EINTR)
	                	goto __ugly2;
                        if (errno == EWOULDBLOCK) 
                                return;  /* no more messages */
	            	log_sys(3024, "recvfrom");
	        }

	        switch ( dh6msg->msg.type ) {
	            	case DH6_REQUEST: 	  
				svr_request_rcv(sia, &msg, len);
				break;
	            	case DH6_RELEASE: 
				svr_release_rcv(sia, &msg, len);
				break;
	            	default:
	                	log_warn(3025,"unknown message %u", 
					dh6msg->msg.type);
	        }
	}
	return;
}

/*
 *	On receiving DHC Request message
 */
int svr_request_rcv(struct svr_ifaddr *sia, struct msghdr *msg, int len)
{
        struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
	struct svr_client  *sc;
	struct ext6_addr *addrext;
	DB      *db;

#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("receiving Request message from %s",
                        inet_ntop(AF_INET6, &src_addr->sin6_addr, 16, NULL));
#endif

	if (len < dh6_request_msg_len(&dh6msg->req)) {
		log_warn(3026, "Request msg is not so long as expected");
		svr_reply_send(sia,msg,DH6_REPLY_POOR_FORM,0,NULL);
		return -1;
	}

	if ( !IN6_IS_ADDR_LINKLOCAL(get_request_llocal_addr(&dh6msg->req)) ) {
		log_warn(3027, "lladdr field in Request not link-local");
		return -1;
	}

	if ( IN6_IS_MULTICAST(&sia->sia_addr) ) {
		log_warn(3028, "Request is destined to a multicast address");
		return -1;
	}
	if ( dh6msg->req.Sbit &&
	     memcmp(&sia->sia_addr,
		    get_request_server_addr(&dh6msg->req), 
		    IPV6_ADDR_LEN) ) {
		log_warn(3029, "dest_addr not the server addr in Request");
		return -1;
	}

	/* 
	 *	Producing error code 20 is doubtful because:
	 *	A client may send a Request message for additional
	 *	configuration parameters; the client may send a
	 *	Request to verify the existing parameters in the server;
	 *	the client may send request to extend its lifetime.
	 */

	dh6db_load_sc(sia, get_request_llocal_addr(&dh6msg->req),
			   get_request_agent_addr(&dh6msg->req));
	sc = sc_get(sia, get_request_llocal_addr(&dh6msg->req),
			 get_request_agent_addr(&dh6msg->req));
	addrext = (struct ext6_addr *)ext6_get(EXT6_ADDR, 
			request_ext(&dh6msg->req), 
			len - dh6_request_msg_len(&dh6msg->req));

	/* 
	 *	Cleaning up the database in required for client's request
 	 * 	TODO: other parameters needs to be cleaned by checking
	 *	      with other extensions.
	 *
	 *	Probably the idea here is not good because the server
	 *	delocate resources any way and then allocate the same
	 *	resources from DH6_SVR_INVAL. Any FIXES would be
	 *	greatly appreciated.
	 */
	if (sc != NULL && dh6msg->req.Cbit &&
	    (addrext == NULL || !addrext->Cbit ||
	     memcmp(&sc->sc_cliaddr,get_ext6_addr_client_addr,IPV6_ADDR_LEN))) {
	        DB  *db, *db2;
        	if ( (db=dh6db_open(DH6_SVR_INUSE)) == NULL)
                	log_sys(3030, "dbopen error");
        	if ( (db2=dh6db_open(DH6_SVR_INVAL)) == NULL)
        	        log_sys(3031, "dbopen error");
        
        	if (dh6db_locate(db,sc) == 0) {
#ifdef DH6_DEBUG 
	                if ( look_db_more() )
	       	                 log_info("found the record in cleaning");
#endif                          
                        dh6db_del(db, sc, R_CURSOR);
                        dh6db_sync(db, R_RECNOSYNC);
                        sc->sc_state = SC_STATE_DELETE;
                        dh6db_put(db2, sc, R_NOOVERWRITE);
                        dh6db_sync(db2,R_RECNOSYNC);
                }

        	dh6db_close(db2);
        	dh6db_close(db);

		u_unlink(&sia->sia_clilist, sc);
		sc_destroy(sc);
		sc = NULL;
		if (addrext == NULL) /* just to release all ???? */
			return 0;
	}

	/* 
	 *	Look into address extension and send replies if necessary.
	 * 	What should I do if there is more than one address extension?
	 */
	svr_request_addrext_rcv(sia, &sc, msg, addrext);

	/*
	 *	TODO: Should I look into other extensions and
	 *	      additionally send other replies?
	 *
	 *	Accept some extensions but may reject others.
	 *	See page 22, draft 08.
	 */
	return 0;
}

/*
 *	Handle an Address extension and send DHC Reply if required
 */
int svr_request_addrext_rcv(struct svr_ifaddr *sia, struct svr_client **sc, 
	struct msghdr *msg, struct ext6_addr *addrext)
{
        struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
	void 	*ext = NULL;
	int	extlen = 0;
	DB      *db;

	/*
	 *	Here we are sure that there is no record in DH6_SVR_INUSE.
	 */
	if ( *sc == NULL ) {
		u_int32_t prefer, valid;

		if ( (*sc=sc_create()) == NULL ) {
			svr_reply_send(sia,msg,DH6_REPLY_RSRC_UNAVAIL,0,NULL);
			return -1;
		}
		u_queue_head(&sia->sia_clilist, *sc);
		(*sc)->sc_sia = sia;
		memcpy(&(*sc)->sc_lladdr, 
			get_request_llocal_addr(&dh6msg->req), 
			sizeof(struct in6_addr));
		memcpy(&(*sc)->sc_rlyaddr, 
			get_request_agent_addr(&dh6msg->req),
			sizeof(struct in6_addr));

		if ( !dh6_any_idle_dh6addr() ) {
			svr_reply_send(sia,msg,DH6_REPLY_RSRC_UNAVAIL,0,NULL);
			u_unlink(&sia->sia_clilist,*sc);
			sc_destroy(*sc);
			return -1;
		}
		if (addrext && addrext->Cbit) {
			memcpy(&(*sc)->sc_cliaddr, 
			       get_ext6_addr_client_addr(addrext), 
			       sizeof(struct in6_addr));
			if ( dh6_get_requested_dh6addr(*sc) )
				;
			else if ( addrext->Qbit ) {
				svr_reply_send(sia,msg,
					DH6_REPLY_UNABLE_EXT_PARA,0,NULL);
				u_unlink(&sia->sia_clilist,*sc);
				sc_destroy(*sc);
				return -1;
			}
			else
				dh6_get_idle_dh6addr(*sc);
		} else
			dh6_get_idle_dh6addr(*sc);

		(*sc)->sc_id = ntohs(dh6msg->req.id);

		if ( addrext ) {
			prefer = get_ext6_addr_prefer_lifetime(addrext);
			if (prefer > MaxPreferLife)
				prefer = DH6_DEFAULT_PREFER_TIME;
			valid  = get_ext6_addr_valid_lifetime(addrext);
			if (valid  > MaxValidLife)
				valid  = DH6_DEFAULT_VALID_TIME;
		} else {
			prefer = DH6_DEFAULT_PREFER_TIME;
			valid  = DH6_DEFAULT_VALID_TIME;
		}
#ifdef TODO
		if (is_specific_client(get_request_llocal_addr(&dh6msg->req))) {
			prefer = ~0;
			valid  = ~0;
		}
#endif

		ext = SVR_APPEND_ADDR_EXT6(ext, &extlen, &(*sc)->sc_cliaddr, 
				(*sc)->sc_prfsize, prefer, valid);
			/* no duplication detection required ? */	
		ext = svr_append_multicast_ext6(ext, &extlen, &MulticastAddr);
		svr_reply_send(sia,msg,DH6_REPLY_SUCCESS,extlen,ext);
		free(ext);

		/*
		 * TODO:
		 * If the client requests DNS updates, and sets the 'Q' bit,
		 * the server MUST NOT issue the DHCP Reply until after 
		 * receiving positive indication that the DNS update has 
		 * indeed been informed. (page 7, ext6)
		 */

                (*sc)->sc_xmit_timer.function =
                        (void (*)(void *))sc_xid_timeout;
		start_timer(&(*sc)->sc_xmit_timer, XID_TIMEOUT);
        	(*sc)->sc_xmit_timeout = XID_TIMEOUT;
		(*sc)->sc_state = SC_STATE_TEMP;
		if ( (db=dh6db_open(DH6_SVR_INUSE)) < 0)
			log_sys(3032, "dbopen");
		dh6db_put(db, *sc, R_NOOVERWRITE);
		dh6db_sync(db);
		dh6db_close(db);
	}

	/*
	 * 	This client previously made a request.
	 *	and thus there must be a record in DH6_SVR_INUSE.
	 */
	else { 	
		u_int32_t prefer, valid;
		/*
		 * Go the same way whether or not it is a retransmission:
		 * ( ntohs(dh6msg->req.id) == (*sc)->sc_id )
		 */
		(*sc)->sc_id = ntohs(dh6msg->req.id);

		if ( addrext ) {
			prefer = get_ext6_addr_prefer_lifetime(addrext);
			if (prefer > MaxPreferLife)
				prefer = DH6_DEFAULT_PREFER_TIME;
			valid  = get_ext6_addr_valid_lifetime(addrext);
			if (valid  > MaxValidLife)
				valid  = DH6_DEFAULT_VALID_TIME;
		} else {
			prefer = DH6_DEFAULT_PREFER_TIME;
			valid  = DH6_DEFAULT_VALID_TIME;
		}
#ifdef TODO
		if (is_specific_client(get_request_llocal_addr(&dh6msg->req))) {
			prefer = ~0;
			valid  = ~0;
		}
#endif

		ext = SVR_APPEND_ADDR_EXT6(ext, &extlen, &(*sc)->sc_cliaddr, 
				(*sc)->sc_prfsize, prefer, valid);
			/* no duplication detection required ? */	
		svr_reply_send(sia,msg,DH6_REPLY_SUCCESS,extlen,ext);      
		free(ext);
                (*sc)->sc_xmit_timer.function =
                        (void (*)(void *))sc_xid_timeout;
		start_timer(&(*sc)->sc_xmit_timer, XID_TIMEOUT);
        	(*sc)->sc_xmit_timeout = XID_TIMEOUT;
		(*sc)->sc_state = SC_STATE_TEMP;
		if ( (db=dh6db_open(DH6_SVR_INUSE)) < 0)
			log_sys(3033, "dbopen");
		if (dh6db_locate(db, *sc) == 0)
			dh6db_put(db, *sc, R_CURSOR);
		dh6db_sync(db);
		dh6db_close(db);
	}
	return 0;
}


/*
 *	On receiving DHC Release message
 */
int svr_release_rcv(struct svr_ifaddr *sia, struct msghdr *msg, int len)
{
        struct sockaddr_in6 *src_addr = (struct sockaddr_in6 *)msg->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg->msg_iov[0].iov_base;
       	struct svr_client  *sc;
	struct ext6_addr *addrext;
	DB  *db, *db2;
	int extlen;

#ifdef DH6_DEBUG
        if ( look_xmit() )
                log_info("receiving Release message from %s",
                        inet_ntop(AF_INET6, &src_addr->sin6_addr, 16, NULL));
#endif

        if ( !IN6_IS_ADDR_LINKLOCAL(&dh6msg->rel.lladdr) ) {
		log_warn(3034, "lladdr in Release not link-local");
                return -1;
	}
	addrext = (struct ext6_addr *)ext6_get(EXT6_ADDR, dh6msg->rel.ext,
			len - dh6_release_msg_len(&dh6msg->rel));

	if ( addrext != NULL && addrext->Cbit && dh6msg->rel.Dbit && 
             memcmp(&src_addr->sin6_addr,
		     get_ext6_addr_client_addr(addrext), IPV6_ADDR_LEN) ) {
		svr_reply_send(sia,msg,DH6_REPLY_INVAL_REL_SRC,0,NULL);
		log_warn(3035, "Reply directly sent but source addr mismatch");
		return -1;	/* page 19, draft 08 */
	}

#ifdef TODO
	if (addrext == NULL && there_are_other_extensions(&dh6msg->rel) ) {
		void *ext2 = NULL;
		int   len2 = 0;
		for (;;) {
			ext2 = server_cannot_process(&dh6msg->rel,&len2);
			if (ext2 == NULL || len2 == 0)
				break;
			svr_reply_send(sia,msg,relevant_error,len2,ext2);
		}
		for (;;) {
			ext2 = server_can_process(&dh6msg->rel,&len2);
			if (ext2 == NULL || len2 == 0)
				return 0;
			svr_reply_send(sia,msg,DH6_REPLY_SUCCESS,len2,ext2);
			release_particular_parameters(sia,msg);
			svr_purge_db(sia, msg);
		}
	}
#endif

	extlen = (addrext) ? ext6_addr_len(addrext) : 0;
	if ( addrext ) {
		set_ext6_addr_prefer_lifetime(addrext, 0);
		set_ext6_addr_valid_lifetime(addrext, 0);
	}
	dh6db_load_sc(sia, &dh6msg->rel.lladdr, &dh6msg->rel.AgtAddr);
	sc = sc_get(sia, &dh6msg->rel.lladdr, &dh6msg->rel.AgtAddr);
        if ( sc == NULL ) {
		svr_reply_send(sia,msg,DH6_REPLY_CLI_REC_UNAVAIL, 
				extlen, addrext);
		return -1;
	}
	svr_reply_send(sia, msg, DH6_REPLY_SUCCESS, extlen, addrext);

        if ( (db=dh6db_open(DH6_SVR_INUSE)) == NULL)
                log_sys(3036, "dbopen error");
       	if ( (db2=dh6db_open(DH6_SVR_INVAL)) == NULL)
              	log_sys(3037, "dbopen error");
	if (dh6db_locate(db,sc) == 0) {
		dh6db_del(db, sc, R_CURSOR);
		dh6db_sync(db, R_RECNOSYNC);
		sc->sc_state = SC_STATE_DELETE;
		dh6db_put(db2, sc, R_NOOVERWRITE);
		dh6db_sync(db2,R_RECNOSYNC);
	}
        dh6db_close(db2);
        dh6db_close(db);

	return 0;
}


static void svr_run_unix_server( void * data )
{
        struct fd_list          *p = (struct fd_list *)data, *fdp;
        int                     fd;
        struct sockaddr_un      unix_addr;
        int                     unix_len = sizeof(unix_addr);

        for (;;) {
                fd = accept(p->fd, (struct sockaddr *)&unix_addr, &unix_len);
                if (fd < 0) {
                        if ( errno!=EWOULDBLOCK )
                                log_sys(3038, "accept");
                        else  
                                return;  /* no more pending connections */
                } else {
                        if ( (fdp=malloc(sizeof(struct fd_list))) == NULL )
                                log_sys(3039, "malloc");
                        init_fdp(fdp);
                        fdp->fd   = fd;
                        fdp->data = fdp;
                        fdp->function = svr_adm_msgrcv;
                        add_fdp(fdp);
                }           
        }          
}                               


/*
 *	to receive administration commands
 */
static void svr_adm_msgrcv( void * data )
{
        struct fd_list *p = (struct fd_list *)data;
        ADM_MESSAGE  msg;
        DH6_MESSAGE *dh6=(DH6_MESSAGE *)(&msg.un.dh6.msg);
        int len;

        for (;;) {          
                len = recv(p->fd, (char*)&msg, MAX_MSG_LEN, 0);
                if ( len==0 )
                        return;
                else if (len < 0) {
                        if (errno == EINTR)
                                continue;
                        else
                                log_sys(3040, "recv");
                }

#ifdef DH6_DEBUG
        	if ( look_xmit() )
                        log_info("receiving a message from dh6sc");
#endif
                if (msg.cmd == ADM_CMD_STOP ) {
                        quit_flag = 1;
                        break;
                }
                if (msg.cmd == ADM_SHOW_DB ) {
                        svr_dbshow_adm(msg, len);
                        continue;
                }

                if (msg.cmd == ADM_DH6_MSG ) {
                        switch ( dh6->msg.type ) {
                        case DH6_RECONFIGURE:
                                svr_reconfigure_adm(&msg, len);
                                break;
                        default:
                                log_warn(1128,"unknown message %u",
                                        dh6->msg.type);
                        }
                }
        }
        del_fdp(p); p->fd=-1;
        close(p->fd);
        free(p);
}

/*
 *	On receiving a Reconfigure message from another process
 */
int svr_reconfigure_adm( ADM_MESSAGE *msg, int len )
{
        DH6_MESSAGE *dh6=(DH6_MESSAGE *)(&msg->un.dh6.msg);
        struct svr_ifaddr *sia;
        struct sockaddr_in6  to6;

        sia = sia_get(&msg->un.dh6.laddr);
        if ( sia == NULL ) {
                log_warn(3129, "The server entry not found");
                return -1;
        }

        bzero((char *)&to6, sizeof(to6));
        to6.sin6_family = AF_INET6;
        memcpy(&to6.sin6_addr, &msg->un.dh6.faddr, sizeof(struct in6_addr));
        to6.sin6_port = htons(DH6_CLIENT_PORT);

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Reconfigure len:%d, to: %s, port:%d", 
			msg->len-ADM_DH6MSG_HDRLEN,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_reconfigure2:
	if (sendto(sia->sia_sk.fd, dh6, msg->len-ADM_DH6MSG_HDRLEN, 0, 
		(struct sockaddr *)&to6, sizeof(struct sockaddr_in6)) 
			!= msg->len-ADM_DH6MSG_HDRLEN) {
	        if ( errno==EINTR )
	                goto _resend_reconfigure2;
	        if ( errno==EINVAL ) 
	                goto _resend_reconfigure2;
		log_sys_warn(3046, "fd(%d):sendto", sia->sia_sk.fd);
	}

	return 0;
}

/*
 *	When the XID_TIMEOUT expires
 */
void sc_xid_timeout( void * data )
{
	struct svr_client	*sc = (struct svr_client *)data;
	DB  *db;

	sc->sc_xmit_timeout = 0;
	sc->sc_state = SC_STATE_PREFER;
	if ( (db=dh6db_open(DH6_SVR_INUSE)) < 0)
		log_sys(3043, "dbopen");
	if (dh6db_locate(db, sc) == 0)
		dh6db_put(db, sc, R_CURSOR);
	dh6db_sync(db);
	dh6db_close(db);
}

/*
 *	When DB check timer expires
 */
void db_check_timeout( void * data )
{
	dh6db_check_purge();
	start_timer(&db_check_timer,SVR_DB_CHECK_INTERVAL);
}

int svr_advertise_send( 
	struct svr_ifaddr *sia, 	/* server address entry */
	struct in6_addr	  *lladdr,	/* the link local address of client */
	struct in6_addr	  *agent,	/* relay agent address */
	int		   extlen,	/* length of extensions */
	void 		  *ext)		/* extensions */
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
	if (agent != NULL) {
    		memcpy(&to6.sin6_addr, agent, sizeof(struct in6_addr));
    		to6.sin6_port = htons(DH6_AGENT_PORT);
	}
	else {
    		memcpy(&to6.sin6_addr, lladdr, sizeof(struct in6_addr));
    		to6.sin6_port = htons(DH6_CLIENT_PORT);
	}

    	bzero((char *)msg, MAX_MSG_LEN);
	msg->adv.type = DH6_ADVERTISE;
	if ( IN6_IS_ADDR_LINKLOCAL(&to6.sin6_addr) )
		msg->adv.lifetime = 0;
	else
		msg->adv.lifetime = sia->sia_lifetime;
	if (agent == NULL) {
		msg->adv.Lbit = 0;
		msg->adv.Sbit = 1;
		msg->adv.SvrCnt = 0;
		memcpy(&msg->adv.AgtAddr, &sia->sia_addr, IPV6_ADDR_LEN);
		len = sizeof(struct dh6_advertise);
	} else {
		msg->adv.Lbit = 1;
		msg->adv.Sbit = 0;
		/*
		 *	When can we give more server addresses?
	 	 */
		msg->adv.SvrCnt = 1;
		memcpy(&msg->adv.AgtAddr, agent, IPV6_ADDR_LEN);
		set_advertise_llocal_addr(&msg->adv, lladdr);
		set_advertise_server_addr(&msg->adv, &sia->sia_addr, 1);
		len = sizeof(struct dh6_advertise)
		      + (msg->adv.SvrCnt+1)*IPV6_ADDR_LEN;
	}

	if ( extlen != 0 && ext != NULL ) {
		memcpy(advertise_ext(&msg->adv), ext, extlen);
		len += extlen;
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Advertise len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
        if ( look_xmit_more() )
		log_info("where agent = %s", 
			inet_ntop(AF_INET6, &msg->adv.AgtAddr, 16, NULL));
#endif 

_resend_advertise:
	if (sendto(sia->sia_sk.fd, msg, len, 0, (struct sockaddr *)&to6, 
		sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_advertise;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Advertise");
	                goto _resend_advertise;
		}
		log_warn(3044, "sendto error %d: %s", errno, strerror(errno));
	}
	return 0;
}

int svr_reply_send( 
	struct svr_ifaddr *sia,		/* server entry */
        struct msghdr 	  *msg0,	/* original message */
	int     	   code,	/* error code */
	int		   extlen,	/* length of extensions */
	void 		  *ext)		/* extensions */
{
        struct sockaddr_in6 *sdest = (struct sockaddr_in6 *)msg0->msg_name;
        DH6_MESSAGE  *dh6msg = (DH6_MESSAGE *)msg0->msg_iov[0].iov_base;
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;
	struct in6_addr *lladdr, *agent;

	if (dh6msg->req.type == DH6_REQUEST) {
		lladdr = get_request_llocal_addr(&dh6msg->req);
		agent  = get_request_agent_addr(&dh6msg->req);
	}
	else {
		lladdr = &dh6msg->rel.lladdr;
		agent  = &dh6msg->rel.AgtAddr;
	}
    	bzero((char *)&to6, sizeof(to6));
    	bzero((char *)msg, MAX_MSG_LEN);

    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, &sdest->sin6_addr, IPV6_ADDR_LEN);

	msg->rpl.type = DH6_REPLY;
	msg->rpl.code = code;
	msg->rpl.id   = dh6msg->req.id;

	if ( !memcmp(&sdest->sin6_addr, agent, IPV6_ADDR_LEN) ||
	     (dh6msg->rel.type == DH6_RELEASE && !dh6msg->rel.Dbit)) {
    		to6.sin6_port = htons(DH6_AGENT_PORT);
		msg->rpl.Lbit = 1;
		set_reply_llocal_addr(&msg->rpl, lladdr);
	}
	else {
    		to6.sin6_port = htons(DH6_CLIENT_PORT);
		msg->rpl.Lbit = 0;
	}

	len = sizeof(struct dh6_reply) + msg->rpl.Lbit*IPV6_ADDR_LEN;

	if ( extlen != 0 && ext != NULL ) {
		memcpy(reply_ext(&msg->rpl), ext, extlen);
		len += extlen;
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Reply len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_reply:
	if (sendto(sia->sia_sk.fd, msg, len, 0, 
		(struct sockaddr *)&to6, sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_reply;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Reply");
	                goto _resend_reply;
		}
		log_sys_warn(3045, "fd(%d):sendto", sia->sia_sk.fd);
	}
	return 0;
}

int svr_reconfigure_send( 
	struct svr_ifaddr *sia,		/* server entry */
	struct in6_addr	  *dest,	/* destination address */
	int		   extlen,	/* length of extensions */
	void 		  *ext)		/* extensions */
{
	struct sockaddr_in6  to6;
	u_int8_t	buf[MAX_MSG_LEN];
	DH6_MESSAGE	*msg = (DH6_MESSAGE *)buf;
	int		len;

    	bzero((char *)&to6, sizeof(to6));
    	to6.sin6_family = AF_INET6;
    	memcpy(&to6.sin6_addr, dest, IPV6_ADDR_LEN);
    	to6.sin6_port = htons(DH6_CLIENT_PORT);

    	bzero((char *)msg, MAX_MSG_LEN);
	msg->conf.type = DH6_RECONFIGURE;
	len = sizeof(struct dh6_reconfigure);

	if ( extlen != 0 && ext != NULL ) {
		memcpy(msg->conf.ext, ext, extlen);
		len += extlen;
	}

#ifdef DH6_DEBUG
        if ( look_xmit() )
		log_info("sending DHC Reconfigure len:%d, to: %s, port:%d", len,
			inet_ntop(AF_INET6, &to6.sin6_addr, 16, NULL),
                         ntohs(to6.sin6_port));
#endif 

_resend_reconfigure:
	if (sendto(sia->sia_sk.fd, msg, len, 0, 
		(struct sockaddr *)&to6, sizeof(struct sockaddr_in6)) != len) {
	        if ( errno==EINTR )
	                goto _resend_reconfigure;
	        if ( errno==EINVAL ) {
			log_info("Resending DHC Reconfigure");
	                goto _resend_reconfigure;
		}
		log_sys_warn(3046, "fd(%d):sendto", sia->sia_sk.fd);
	}
	return 0;
}
