/*
 * ext6.c    	Extensions for DHCPv6 
 *
 * Version:     @(#)ext6.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include "dh6.h"
#include "ext6.h"

/* length of address extension */
inline int ext6_addr_len(struct ext6_addr *ext)
{
	return 	sizeof(struct ext6_addr) +
		ext->Cbit * IPV6_ADDR_LEN +
#ifdef TODO
		strlen(get_ext6_addr_dns_name(ext)) + 1 +
#endif
		ext->Lbit * sizeof(u_int32_t) * 2;
}

/* client address if present */
inline struct in6_addr * 
get_ext6_addr_client_addr( struct ext6_addr *ext )
{
	if ( ext->Cbit )
		return( (struct in6_addr *) (ext->data) );
	else
		return NULL;
}

inline void
set_ext6_addr_client_addr( struct ext6_addr *ext, struct in6_addr *cliaddr )
{
	if ( !ext->Cbit )
		return;
	memcpy(ext->data, cliaddr, IPV6_ADDR_LEN);
}

/* preferred lifetime if present */
inline u_int32_t
get_ext6_addr_prefer_lifetime( struct ext6_addr *ext )
{
	if ( ext->Lbit )
		return( *((u_int32_t *)(ext->data + 
			ext->Cbit*sizeof(struct in6_addr))) );
	else
		return 0;
}

inline void
set_ext6_addr_prefer_lifetime( struct ext6_addr *ext, u_int32_t lifetime )
{
	if ( !ext->Lbit )
		return;
	*((u_int32_t *)(ext->data + ext->Cbit*sizeof(struct in6_addr))) 
		= lifetime;
}

/* valid lifetime if present */
inline u_int32_t
get_ext6_addr_valid_lifetime( struct ext6_addr *ext )
{
	if ( ext->Lbit )
		return( *((u_int32_t *)(ext->data + 
				ext->Cbit*sizeof(struct in6_addr) + 
				ext->Lbit*sizeof(u_int32_t))) );
	else
		return  0;
}

inline void 
set_ext6_addr_valid_lifetime( struct ext6_addr *ext, u_int32_t lifetime )
{
	if ( !ext->Lbit )
		return;
	*((u_int32_t *)(ext->data + ext->Cbit*sizeof(struct in6_addr) + 
			ext->Lbit*sizeof(u_int32_t)))  = lifetime;
}

/* DNS name if present */
inline char *
get_ext6_addr_dns_name( struct ext6_addr *ext )
{
	return( (char *)(ext->data + ext->Cbit*sizeof(struct in6_addr) + 
				ext->Lbit*sizeof(u_int32_t)*2) );
}

inline void 
set_ext6_addr_dns_name( struct ext6_addr *ext, char *dns )
{
	strcpy( (char *)(ext->data + ext->Cbit*sizeof(struct in6_addr) + 
		ext->Lbit*sizeof(u_int32_t)*2), dns );
}



/* directory agent address if present */
inline struct in6_addr * 
get_ext6_da_addr( struct ext6_da *ext )
{
	if ( ext->Dbit )
		return( (struct in6_addr *) (ext->data) );
	else
		return NULL;
}

inline void
set_ext6_da_addr( struct ext6_da *ext, struct in6_addr *addr )
{
	if ( !ext->Dbit )
		return;
	memcpy(ext->data, addr, IPV6_ADDR_LEN);
}

/* Scope list: if there is nothing in Scope list, does this list include only '\0'? */
char ** get_ext6_da_scope_list( struct ext6_da *ext)
{
	static char **p=NULL;
	register int i;
	p = (char **)realloc( p, ext->ScopeCnt*sizeof(char *) );
	p[0] = ext->data + ext->Dbit*sizeof(struct in6_addr);
	for (i=1; i<ext->ScopeCnt; i++)
		p[i] = p[i-1] + strlen(p[i-1]) + 1;
	return( p );
}

void set_ext6_da_scope_list( struct ext6_da *ext, char *list[] )
{
	char *p;
	register int i, pos = 0;
	p = ext->data + ext->Dbit*sizeof(struct in6_addr);
	for (i=1; i<ext->ScopeCnt; i++) {
		strcpy(p+pos, list[i]);
		pos += strlen(list[i]) + 1;
	}
}

/* destination addresses */
inline struct in6_addr * 
get_ext6_static_route_dest( struct ext6_static_route *ext, int seqno )
{
	if ( (seqno+1)*32 > ext->len )
		return NULL;
	return( (struct in6_addr *) (ext->a[seqno]) );
}

inline void
set_ext6_static_route_dest( struct ext6_static_route *ext, 
	int seqno, struct in6_addr *dest )
{
	if ( (seqno+1)*32 > ext->len )
		return;
	memcpy(ext->a[seqno], dest, IPV6_ADDR_LEN);
}

/* router addresses */
inline struct in6_addr * 
get_ext6_static_route_router( struct ext6_static_route *ext, int seqno )
{
	if ( (seqno+1)*32 > ext->len )
		return NULL;
	return( (struct in6_addr *) (ext->a[seqno] + sizeof(struct in6_addr)) );
}

inline void
set_ext6_static_route_router( struct ext6_static_route *ext, 
	int seqno, struct in6_addr *router )
{
	if ( (seqno+1)*32 > ext->len )
		return;
	memcpy(ext->a[seqno]+IPV6_ADDR_LEN, router, IPV6_ADDR_LEN);
}

/* locate a specific extension */
u_int8_t * 
ext6_get(int code, void *ext6, int len)
{
	u_int8_t 	*ext = (u_int8_t *)ext6, 
			*end = ext + len - 1;

	while (ext <= end) {
		int code2 = ntohs(*((u_int16_t *)ext));
		if (code2 == code)
			return ext;
		if (code2 == EXT6_END)
			return NULL;
		if (ext+1 == end && *ext == EXT6_PAD)
			return NULL;
		ext += 4 + ntohs(*((u_int16_t *)(ext+2)));
	}
	return NULL;
}

void * append_addr_ext6(
    void 	      *ext, 		/* original extensions */
    int 	      *len,		/* length of original extensions */
    struct in6_addr   *cliaddr,		/* client address */ 
    int		       PrfSize,		/* prefix size */
    u_int32_t	       prefer,		/* preferred lifetime */
    u_int32_t	       valid,		/* valid lifetime */
    char	      *dns,		/* DNS Name */
    u_int8_t	       Qbit,		/* Required bit */
    u_int8_t	       Abit,		/* AAAA bit */
    u_int8_t	       Pbit)		/* new PTR bit */
{
	struct ext6_addr *p;
	void 	*ext2;
	void    *tmp = malloc(*len);
	int	newlen = sizeof(struct ext6_addr);

	memcpy(tmp, ext, *len);

	if ( cliaddr )
		newlen += IPV6_ADDR_LEN;
	if (prefer != 0)
		newlen += 2 * sizeof(u_int32_t);
	if (dns != NULL)
		newlen += strlen(dns) + 1;
	if ( ext )
		ext2 = realloc(ext, *len+newlen);
	else
		ext2 = malloc(*len+newlen);
	memcpy(ext2, tmp, *len);
	free(tmp);

	p = (struct ext6_addr *)( ((u_int8_t *)ext2) + *len );
	*len += newlen;

	p->code = htons(EXT6_ADDR);
	p->len  = htons(newlen - 4);
	p->Cbit = (cliaddr != NULL);
	p->Lbit = (prefer != 0);
	p->Qbit = Qbit;
	p->Abit = Abit;
	p->Pbit = Pbit;
	p->PrfSize = PrfSize;

	if ( cliaddr )
		set_ext6_addr_client_addr(p, cliaddr);
	if ( p->Lbit ) {
		set_ext6_addr_prefer_lifetime(p, prefer);
		set_ext6_addr_valid_lifetime(p, valid);
	}
	if ( dns )
		set_ext6_addr_dns_name(p, dns);
	
	return ext2;
}

void * svr_append_multicast_ext6(
    void 	      *ext, 		/* original extensions */
    int 	      *len,		/* length of original extensions */
    struct in6_addr   *addr)		/* multicast address */ 
{
	struct ext6_multicast *p;
	void 	*ext2;
	void    *tmp = malloc(*len);
	int	newlen = sizeof(struct ext6_multicast);

	memcpy(tmp, ext, *len);
	if ( ext )
		ext2 = realloc(ext, *len+newlen);
	else
		ext2 = malloc(*len+newlen);
	memcpy(ext2, tmp, *len);
	free(tmp);

	p = (struct ext6_multicast *)( ((u_int8_t *)ext2) + *len );
	*len += newlen;

	p->code = htons(EXT6_MULTICAST);
	p->len  = htons(newlen - 4);
	memcpy(&p->addr, addr, IPV6_ADDR_LEN);

	return ext2;
}
