/*
 * parse.c    	Parsing Routines for DHCPv6 Configuration Files
 *
 * Version:     @(#)parse.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include <db.h>
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dispatch.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_adm.h"
#include "parse.h"

char *prev_line;
char *curr_line;

static char line[2][MAX_LINE_LEN];
static int ln_pos;
static int ln_num;

static inline int psdeo_line( int lno )
{
	return( lno % 2 );
}

static int parse_getchar( FILE *f )
{
	int c = getc(f);
	if (c == '\n') {
		prev_line = line[psdeo_line(ln_num)];
		curr_line = line[psdeo_line(++ln_num)];
		ln_pos = 0;
		curr_line[0] = 0;
	} else if (c != EOF) {
		if (ln_pos < MAX_LINE_LEN ) {
			curr_line[ln_pos] = c;
			curr_line[ln_pos+1] = 0;
		}
		ln_pos++;
	}
	return c;		
}

static inline void skip_to_eol( FILE *f )
{
	int c;
	while ( (c=parse_getchar(f)) != EOF ) {
		if (c == '\n')
			return;
	} 
}

/*
 *	a string is identified by "..."
 */
static int extract_string( FILE *f, char *buf )
{
	int i, c;
	int continued = 0;

	for (i = 0; i < MAX_TOKEN_LEN; i++) {
		c = parse_getchar(f);
		if (c == EOF) {
			log_warn("eof in string token");
			break;
		}
		if (continued) {
			continued = 0;
			buf[i] = c;
		} else if (c == '\\')
			continued = 1;
		else if (c == '"')
			break;
		else
			buf[i] = c;
	}
	if (i == MAX_TOKEN_LEN) {
		log_warn("string token too long");
		--i;
	}
	buf[i] = 0;
	return STRING;
}

static int extract_number( int first, FILE *f, char *buf )
{
	int xflag = 0;
	int i=0, c;
	int type = DIGIT;

	buf[i++] = first;
	for (; i < MAX_TOKEN_LEN; i++) {
		c = parse_getchar(f);
		if ( isascii(c) && isspace(c) )
			break;
		else if (!xflag && (c == 'x' || c == 'X')) {
			/* allow one 'x' only, such as 0x3d */
			xflag = 1;
			type |= XPREFIX;
		} else if (isascii(c) && !isxdigit(c) &&
			   (c == '-' || c == '_' || isalpha (c))) {
			type |= _ALPHA;
		} else if ( isdigit(c) ) {
			type |= DIGIT;
		} else if (isascii(c) && isxdigit(c)) {
			type |= XDIGIT;
		} else if (!isascii(c) || !isxdigit(c)) {
			type |= UNKNOWN;
			break;
		}
		buf[i] = c;
	}
	if (i == MAX_TOKEN_LEN) {
		log_warn("numeric token too long");
		--i;
	}
	else
		ungetc(c, f);
	buf[i] = 0;
	return type;
}

static int extract_identifier( int first, FILE *f, char *buf )
{
	int i = 0, c;
	int type = NONE;

	buf [i++] = first;
	for (; i < MAX_TOKEN_LEN; i++) {
		c = parse_getchar(f);
		if ( isascii(c) && isspace(c) )
			break;
		if (!isascii(c) ||
		    (c != '-' && c != '_' && !isalnum (c))) {
			ungetc(c, f);
			type |= UNKNOWN;
			break;
		}
		if (isdigit(c))
			type |= DIGIT;
		else if (isxdigit(c))
			type |= XDIGIT;
		else if (isascii(c))
			type |= _ASCII;
		else if (isalpha(c))
			type |= _ALPHA;
		buf[i] = c;
	}
	if (i == MAX_TOKEN_LEN) {
		log_warn("token too long");
		--i;
	}
	else
		ungetc(c, f);
	buf[i] = 0;

	return seek_match(buf, type);
}

void parse_init( void )
{
	ln_pos = 0;
	ln_num = 0;
	curr_line = line[0];
	prev_line = line[1];
	curr_line[0] = prev_line[0] = 0;
}

int get_next_token( FILE *f, char *buf )
{
	int c;
	int type = NONE;

	do {
		c = parse_getchar(f);
		if (isascii(c) && isspace(c))
			continue;
		if (c == '#') {
			skip_to_eol(f);
			continue;
		}
		if (c == '"') {
			type = extract_string(f, buf);
			break;
		}
		if ((isascii(c) && isdigit(c)) || c == '-') {
			type = extract_number(c, f, buf);
			break;
		} else if (isascii(c) && isalpha(c)) {
			type = extract_identifier(c, f, buf);
			break;
		} else {
			type = c;
			break;
		}
	} while (1);
#ifdef DH6_DEBUG
        if ( look_db_more() )
		log_info("type = %04X or %d, buf = %s", type, type, buf);
#endif
	return type;
}

int seek_identifer( int type, FILE *f )
{
        char buf[100];
	int type2;

	rewind(f);
	parse_init();
	do {
                type2 = get_next_token(f, buf);
                if (type2 == type)
                        return type;
        } while (type2 != EOF);
        return EOF;
}

int fcopy_until(int type, FILE *to, FILE *from)
{
	int c;
        char buf[100];
	int type2;

#ifdef DH6_DEBUG
        if ( look_db_more() )
		log_info("fcopy_until()");
#endif
	for (;;) {
		c = parse_getchar(from);
        	if (c == EOF)
			break;
		if ( isspace(c)) {
			fputc(c, to);
			continue;
		}
		if (c == '#') {
			fputc('#', to);
			fgets(buf, MAX_LINE_LEN, from);
			fputs(buf, to);
			continue;
		}
		if (c == '"') {
			extract_string(from, buf);
			fprintf(to, "\"%s\"", buf);
#ifdef DH6_DEBUG
        		if ( look_db_more() )
				log_info("buf:%s", buf);
#endif
			continue;
		}
		if ((isascii(c) && isdigit(c)) || c == '-') {
			extract_number(c, from, buf);
			fprintf(to, "%s", buf);
#ifdef DH6_DEBUG
        		if ( look_db_more() )
				log_info("buf:%s", buf);
#endif
			continue;
		} else if (isascii(c) && isalpha(c)) {
			type2 = extract_identifier(c, from, buf);
			if (type2 == type)
				return type;
			fprintf(to, "%s", buf);
			continue;
		} else {
			fputc(c, to);
			continue;
		}
        }
	return EOF;
}
