/*
 * u_buf.c    	DHCPv6 Buffering
 *
 * Version:     @(#)u_buf.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include "compat.h"
#include "localdefs.h"
#include "u_timer.h"
#include "u_buf.h"
#include "dispatch.h"
#include "dh6.h"
#include "ext6.h"
#include "dh6_if.h"
#include "dh6_cli.h"
#include "dh6_adm.h"

#define CHAIN_PTR(p)		((struct chain *)p)
#define LARGE_CHAIN_PTR(p)	((struct large_chain *)p)

/*
 *	table buffer list
 */
static struct chain 		u_buf_base[MAX_TABLES];
static struct large_chain 	u_large_buf_base[MAX_LARGE_TABLES];
UNION_LIST	u_free_list;
UNION_LIST	u_large_free_list;

int u_free_list_init( void )
{
	int i;
	struct chain *p = (struct chain *)u_buf_base;

	p[0].prev = NULL;
	for (i=0; i<MAX_TABLES-1; i++) {
		p[i].next   = p+i+1;
		p[i+1].prev = p+i;
	}
	p[MAX_TABLES-1].next = NULL;
	u_free_list.chain.head = p;
	u_free_list.chain.tail = p+MAX_TABLES-1;

	p = (struct chain *)u_large_buf_base;
	p[0].prev = NULL;
	for (i=0; i<MAX_LARGE_TABLES-1; i++) {
		p[i].next   = p+i+1;
		p[i+1].prev = p+i;
	}
	p[MAX_LARGE_TABLES-1].next = NULL;
	u_large_free_list.chain.head = p;
	u_large_free_list.chain.tail = p+MAX_LARGE_TABLES-1;

	return 0;
}

/*
 *	allocate a chain node
 */

void * u_alloc_node( void )
{
	struct chain *node=u_free_list.chain.tail;

	if ( node==NULL ) {
		log_warn(301, "No more buffer in the free list");
		return NULL;
	}
	u_free_list.chain.tail = node->prev;

	if ( u_free_list.chain.tail==NULL )
		u_free_list.chain.head = NULL;
	else
		u_free_list.chain.tail->next = NULL;
	memset(node, 0, sizeof(struct chain));
	return (void *)node;
}

/*
 *	free a chain node
 */

void u_free_node( void * node )
{
	CHAIN_PTR(node)->prev = u_free_list.chain.tail;
	CHAIN_PTR(node)->next = NULL;

	if ( u_free_list.chain.head==NULL )
		u_free_list.chain.head = CHAIN_PTR(node);
	else
		u_free_list.chain.tail->next = CHAIN_PTR(node);
	u_free_list.chain.tail = CHAIN_PTR(node);
}

/*
 *	allocate a large chain node
 */

void * u_alloc_large_node( void )
{
	struct large_chain *node=u_large_free_list.large_chain.tail;

	if ( node==NULL ) {
		log_warn(302, "No more buffer in the large free list");
		return NULL;
	}
	u_large_free_list.large_chain.tail = node->prev;

	if ( u_large_free_list.large_chain.tail==NULL )
		u_large_free_list.large_chain.head = NULL;
	else
		u_large_free_list.large_chain.tail->next = NULL;
	memset(node, 0, sizeof(struct large_chain));
	return (void *)node;
}

/*
 *	free a large chain node
 */

void u_free_large_node( void * node )
{
	LARGE_CHAIN_PTR(node)->prev = u_large_free_list.large_chain.tail;
	LARGE_CHAIN_PTR(node)->next = NULL;

	if ( u_large_free_list.large_chain.head==NULL )
		u_large_free_list.large_chain.head = LARGE_CHAIN_PTR(node);
	else
		u_large_free_list.large_chain.tail->next = LARGE_CHAIN_PTR(node);
	u_large_free_list.large_chain.tail = LARGE_CHAIN_PTR(node);
}

void * u_clone_node( void *old, int len )
{
	struct chain	*q;
	if ( (q = u_alloc_node()) == NULL )
		return NULL;
	memcpy(q->__, ((struct chain *)old)->__, 
		(len>MAX_SIZE_TABLE)?MAX_SIZE_TABLE:len);
	return q;
}

void * u_clone_large_node( void *old, int len )
{
	struct large_chain	*q;
	if ( (q = u_alloc_large_node()) == NULL )
		return NULL;
	memcpy(q->____, ((struct large_chain *)old)->____, 
		(len>MAX_SIZE_LARGE_TABLE)?MAX_SIZE_LARGE_TABLE:len);
	return q;
}

/*
 *	Append an element to the beginning of a list.
 */

void u_queue_head(UNION_LIST *q, void *new)
{
	CHAIN_PTR(new)->next = q->chain.head;
	CHAIN_PTR(new)->prev = NULL;

	if ( q->chain.tail==NULL )
		q->chain.tail = CHAIN_PTR(new);
	else
		q->chain.head->prev = CHAIN_PTR(new);
	q->chain.head = CHAIN_PTR(new);
}

/*
 *	Insert an element before a chained element in the middle of a list.
 */

void u_fore_queue_mid(UNION_LIST *q, void *mid, void *new)
{
	CHAIN_PTR(new)->next = CHAIN_PTR(mid);
	CHAIN_PTR(new)->prev = CHAIN_PTR(mid)->prev;

	if ( CHAIN_PTR(mid)->prev==NULL )
		q->chain.head = CHAIN_PTR(new);
	else
		CHAIN_PTR(mid)->prev->next = CHAIN_PTR(new);

	CHAIN_PTR(mid)->prev = CHAIN_PTR(new);
}

/*
 *	Insert an element after a chained element in the middle of a list.
 */

void u_hind_queue_mid(UNION_LIST *q, void *mid, void *new)
{
	CHAIN_PTR(new)->prev = CHAIN_PTR(mid);
	CHAIN_PTR(new)->next = CHAIN_PTR(mid)->prev;

	if ( CHAIN_PTR(mid)->next==NULL )
		q->chain.tail = CHAIN_PTR(new);
	else
		CHAIN_PTR(mid)->next->prev = CHAIN_PTR(new);

	CHAIN_PTR(mid)->next = CHAIN_PTR(new);
}

/*
 *	Append an element to the end of a list.
 */

void u_queue_tail(UNION_LIST *q, void *new)
{
	CHAIN_PTR(new)->prev = q->chain.tail;
	CHAIN_PTR(new)->next = NULL;

	if ( q->chain.head==NULL )
		q->chain.head = CHAIN_PTR(new);
	else
		q->chain.tail->next = CHAIN_PTR(new);
	q->chain.tail = CHAIN_PTR(new);
}

/*
 *	Remove an element from a list. Works even without knowing the list it
 *	is sitting on, which can be handy at times. 
 */

void u_unlink(UNION_LIST *q, void *old)
{ 			
	struct chain 	*save_prev;

	if ( !CHAIN_PTR(old)->next )
		q->chain.tail = CHAIN_PTR(old)->prev;
	if ( !CHAIN_PTR(old)->prev )
		q->chain.head = CHAIN_PTR(old)->next;
	save_prev = CHAIN_PTR(old)->prev;
	if ( CHAIN_PTR(old)->prev ) {
		CHAIN_PTR(old)->prev->next = CHAIN_PTR(old)->next;
		CHAIN_PTR(old)->prev = NULL;
	}
	if ( CHAIN_PTR(old)->next ) {
		CHAIN_PTR(old)->next->prev = save_prev;
		CHAIN_PTR(old)->next = NULL;
	}
}

