/*
 * u_time.c    	DHCPv6 Timing
 *
 * Version:     @(#)u_timer.c   1.0     Dec-8-96
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed WITHOUT ANY WARRANTY and even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Many thanks to Jim Bound in Digital Equipment Corp.
 * Appreciate any further work on this release.
 *
 * Authors:     Yunzhou Li (yunzhou@iol.unh.edu)
 *
 */

#include	<stdio.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	<signal.h>
#include	"u_timer.h"

extern int debug;

static long faketime=0;

static int timer_table_lock_flag = 0;

inline void lock_timer_table( void )
{
	timer_table_lock_flag = 1;
}

inline void unlock_timer_table( void )
{
	timer_table_lock_flag = 0;
}

inline int timer_table_locked( void )
{
	return timer_table_lock_flag;
}

void begin_timing( void )
{
    	struct itimerval itval;

   	itval.it_interval.tv_sec = TIMER_INTERVAL/1000;
	itval.it_value.tv_sec = TIMER_INTERVAL/1000;
    	itval.it_interval.tv_usec = (TIMER_INTERVAL%1000)*1000;
    	itval.it_value.tv_usec = (TIMER_INTERVAL%1000)*1000;
	setitimer(ITIMER_REAL, &itval, (struct itimerval *)NULL);
	signal(SIGALRM, run_timer_table);
}

void end_timing( void )
{
    	struct itimerval itval;

	signal(SIGALRM, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);
   	itval.it_interval.tv_sec = 0;
    	itval.it_value.tv_sec = 0;
    	itval.it_interval.tv_usec = 0;
    	itval.it_value.tv_usec = 0;
	setitimer(ITIMER_REAL, &itval, (struct itimerval *)NULL);
}

static struct u_timer timer_head = 
	{ &timer_head, &timer_head, ~0, 0, NULL };

int reset_timer_table( void )
{
	timer_head.prev = &timer_head;
	timer_head.next = &timer_head;
	timer_head.expires = ~0;
	timer_head.data = NULL;
	timer_head.function = NULL;
}

void init_timer(struct u_timer * timer)
{
        timer->next = NULL;
        timer->prev = NULL;
}

static void add_timer(struct u_timer * timer)
{
	struct u_timer *p;

	p = &timer_head;
	do {
		p = p->next;
	} while (timer->expires > p->expires);

#ifdef DH6_DEBUG
       	if ( look_tm() )
		log_info("pid(%d)-add_timer(): %ld", 
			getpid(), timer->expires - faketime);
#endif

	lock_timer_table();
	timer->next = p;
	timer->prev = p->prev;
	p->prev = timer;
	timer->prev->next = timer;
	unlock_timer_table();
}

static int del_timer(struct u_timer * timer)
{
	if (timer->next) {
		struct u_timer * next;

#ifdef DH6_DEBUG
       	if ( look_tm() )
		log_info("pid(%d)-del_timer(): %ld", 
			getpid(), timer->expires - faketime);
#endif

		if ((next = timer->next) != NULL) {
			lock_timer_table();
			(next->prev = timer->prev)->next = next;
			timer->next = timer->prev = NULL;
			unlock_timer_table();
			return (timer->expires - faketime);
		}
	}
	return -1;
}

static struct poll_list poll_head = 
	{ &poll_head, &poll_head, 0, NULL };

int reset_poll_list( void )
{
	poll_head.prev = &poll_head;
	poll_head.next = &poll_head;
	poll_head.data = NULL;
	poll_head.function = NULL;
}

void init_poll(struct poll_list * pl)
{
        pl->next = NULL;
        pl->prev = NULL;
}

/*
 *	add to the tail of the polling list
 */
void add_poll(struct poll_list * pl)
{
	pl->next = &poll_head;
	pl->prev = poll_head.prev;
	poll_head.prev = pl;
	pl->prev->next = pl;
}

void del_poll(struct poll_list * pl)
{
	if (pl->next) {
		struct poll_list * next;
		if ((next = pl->next) != NULL) {
			(next->prev = pl->prev)->next = next;
			pl->next = pl->prev = NULL;
			return;
		}
	}
}

/*
 * 	LOCK YOUR TIMER TABLE:
 *
 *	When you are handling some other interrupts, you should
 *	ALWAYS lock your timer table. Otherwise, certain important
 *	queue might be crashed if your timeout function does some 
 *	critical stuffs related to this queue.
 */
void run_timer_table( int signo )
{
	struct u_timer *t;
	static intvl = 0;

#ifdef DH6_DEBUG
       	if ( look_tm_more() ) 
		log_info("pid(%d)-run_timer_table()", getpid());
#endif

	/*
	 *	What will happen if faketime overflows?
	 */
	faketime++;

	if ( (++intvl) >= 1*SECONDS) {
		struct poll_list *pl;
		pl = poll_head.next;
		while (pl != &poll_head) {
			void (*fn)(void *) = pl->function;
			void *data = pl->data;
			fn(data);
			pl = pl->next;
		}
		intvl = 0;
	}

	if ( timer_table_locked() ) return;

	while ((t = timer_head.next) != &timer_head && 
		    t->expires <= faketime) {
		void (*fn)(void *) = t->function;
		void *data = t->data;

		lock_timer_table();
		t->next->prev = t->prev;
		t->prev->next = t->next;
		t->next = t->prev = NULL;
		unlock_timer_table();

		fn(data);
	}
}

inline int stop_timer( struct u_timer *t )
{
	return del_timer( t );
}

inline void start_timer(struct u_timer *t, u_int16_t len)
{
	t->expires = faketime+len;
	add_timer( t );
}

inline int restart_timer(struct u_timer *t, u_int16_t len)
{
	int ret = stop_timer(t);
	start_timer(t,len);
	return ret;
}

/* ignore a timeout */
inline void u_ignore( void *data )
{ ; }

#ifdef NO_PROBLEM

/*
 *	Please alway use u_sleep() instead of sleep()
 *	because otherwise your timers won't work.
 */
inline unsigned u_sleep(unsigned seconds)
{
	unsigned ret=sleep(seconds);
	signal(SIGALRM, run_timer_table);
	return ret;
}

/*
 *	Please alway use u_alarm() instead of alarm()
 *	because otherwise your timers won't work.
 */
inline unsigned u_alarm(unsigned seconds)
{
	unsigned ret=alarm(seconds);
	signal(SIGALRM, run_timer_table);
	return ret;
}

#endif
